#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""The MacroServer tango module"""

import os.path
import sys

from PyTango import Util, Except, DevVoid, DevLong, DevString, DevState, \
    DevEncoded, DevVarStringArray, READ, READ_WRITE, SCALAR, SPECTRUM, DebugIt

import taurus
from taurus.core.util.codecs import CodecFactory

from sardana import State, SardanaServer
from sardana.tango.core.SardanaDevice import SardanaDevice, SardanaDeviceClass
from sardana.macroserver.msexception import MacroServerException
from sardana.macroserver.macroserver import MacroServer as MS


class MacroServer(SardanaDevice):
    """The MacroServer tango class"""

    ElementsCache = None
    EnvironmentCache = None

    def __init__(self, cl, name):
        self._macro_server = None
        SardanaDevice.__init__(self, cl, name)

    def init(self, name):
        SardanaDevice.init(self, name)

        if self._alias is None:
            self._alias = Util.instance().get_ds_inst_name()

        self._macro_server = ms = MS(self.get_full_name(), self.alias)
        ms.add_listener(self.on_macro_server_changed)

    @property
    def macro_server(self):
        return self._macro_server

    def delete_device(self):
        SardanaDevice.delete_device(self)
        self._macro_server.clear_log_report()
        # Workaround for bug #494.
        factory = taurus.Factory("tango")
        for attr in list(factory.tango_attrs.values()):
            attr.cleanUp()

    def init_device(self):
        SardanaDevice.init_device(self)
        self.set_change_event('State', True, False)
        self.set_change_event('Status', True, False)
        self.set_change_event('TypeList', True, False)
        self.set_change_event('DoorList', True, False)
        self.set_change_event('MacroList', True, False)
        self.set_change_event('MacroLibList', True, False)
        self.set_change_event('Elements', True, False)
        self.set_change_event('Environment', True, False)

        dev_class = self.get_device_class()
        self.get_device_properties(dev_class)

        self.EnvironmentDb = self._calculate_name(self.EnvironmentDb)
        self.LogReportFilename = self._calculate_name(self.LogReportFilename)

        macro_server = self.macro_server
        macro_server.set_python_path(self.PythonPath)
        macro_server.set_max_parallel_macros(self.MaxParallelMacros)

        # if it is not possible to store/retrieve the environment from the
        # current path then setup a new unique path and store the environment
        # there forever
        try:
            macro_server.set_environment_db(self.EnvironmentDb)
        except:
            self.error("Failed to set environment DB to %s",
                       self.EnvironmentDb)
            self.debug("Details:", exc_info=1)
            import tempfile
            env_db = os.path.join(tempfile.mkdtemp(),
                                  MacroServerClass.DefaultEnvRelDir)
            env_db = self._calculate_name(env_db)
            db = Util.instance().get_database()
            db.put_device_property(self.get_name(), dict(EnvironmentDb=env_db))
            self.EnvironmentDb = env_db
            macro_server.set_environment_db(self.EnvironmentDb)

        try:
            macro_server.set_log_report(
                self.LogReportFilename, self.LogReportFormat)
        except:
            self.error("Failed to setup log report to %s",
                       self.LogReportFilename)
            self.debug("Details:", exc_info=1)

        macro_server.set_recorder_path(self.RecorderPath)
        macro_server.set_macro_path(self.MacroPath)
        macro_server.set_pool_names(self.PoolNames)

        if self.RConsolePort:
            try:
                import rfoo.utils.rconsole
                rfoo.utils.rconsole.spawn_server(port=self.RConsolePort)
            except Exception:
                self.warning("Failed to start rconsole")
                self.debug("Details:", exc_info=1)
        self.set_state(DevState.ON)

    def _calculate_name(self, name):
        if name is None:
            return None
        util = Util.instance()
        return name % {'ds_name': util.get_ds_name().lower(),
                       'ds_exec_name': util.get_ds_exec_name(),
                       'ds_inst_name': util.get_ds_inst_name().lower()}

    def on_macro_server_changed(self, evt_src, evt_type, evt_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        evt_name = evt_type.name.lower()

        multi_attr = self.get_device_attr()
        elems_attr = multi_attr.get_attr_by_name("Elements")
        if evt_name == "poolelementschanged":
            # force the element list cache to be rebuild next time someone reads
            # the element list
            self.ElementsCache = None
            self.set_attribute(elems_attr, value=evt_value.rvalue)
            #self.push_change_event('Elements', *evt_value.value)
        elif evt_name in ("elementcreated", "elementdeleted"):
            # force the element list cache to be rebuild next time someone reads
            # the element list
            self.ElementsCache = None

            elem = evt_value

            value = {}
            if "created" in evt_name:
                key = 'new'
            else:
                key = 'del'
            json_elem = elem.serialize(pool=self.pool.full_name)
            value[key] = json_elem,
            value = CodecFactory().getCodec('utf8_json').encode(('', value))
            self.set_attribute(elems_attr, value=value)
            #self.push_change_event('Elements', *value)
        elif evt_name == "elementschanged":
            # force the element list cache to be rebuild next time someone reads
            # the element list
            self.ElementsCache = None

            ms_name = self.macro_server.full_name
            new_values, changed_values, deleted_values = [], [], []
            for elem in evt_value['new']:
                json_elem = elem.serialize(macro_server=ms_name)
                new_values.append(json_elem)
            for elem in evt_value['change']:
                json_elem = elem.serialize(macro_server=ms_name)
                changed_values.append(json_elem)
            for elem in evt_value['del']:
                json_elem = elem.serialize(macro_server=ms_name)
                deleted_values.append(json_elem)
            value = {"new": new_values, "change": changed_values,
                     "del": deleted_values}
            value = CodecFactory().getCodec('utf8_json').encode(('', value))
            self.set_attribute(elems_attr, value=value)
            #self.push_change_event('Elements', *value)
        elif evt_name == "environmentchanged":
            self.EnvironmentCache = None
            env_attr = multi_attr.get_attr_by_name("Environment")
            value = CodecFactory().getCodec('pickle').encode(('', evt_value))
            self.set_attribute(env_attr, value=value)

    def always_executed_hook(self):
        pass

    def read_attr_hardware(self, data):
        pass

    def read_DoorList(self, attr):
        door_names = self.macro_server.get_door_names()
        attr.set_value(door_names)

    @DebugIt()
    def read_MacroList(self, attr):
        macro_names = self.macro_server.get_macro_names()
        attr.set_value(macro_names)

    def read_MacroLibList(self, attr):
        macro_lib_names = self.macro_server.get_macro_lib_names()
        attr.set_value(macro_lib_names)

    def read_TypeList(self, attr):
        type_names = self.macro_server.get_data_type_names_with_asterisc()
        attr.set_value(type_names)

    #@DebugIt()
    def getElements(self, cache=True):
        value = self.ElementsCache
        if cache and value is not None:
            return value
        elements = self.macro_server.get_elements_info()
        value = dict(new=elements)
        value = CodecFactory().getCodec('utf8_json').encode(('', value))
        self.ElementsCache = value
        return value

    #@DebugIt()
    def read_Elements(self, attr):
        fmt, data = self.getElements()
        attr.set_value(fmt, data)

    def is_Elements_allowed(self, req_type):
        return SardanaServer.server_state == State.Running

    is_DoorList_allowed = \
        is_MacroList_allowed = \
        is_MacroLibList_allowed = \
        is_TypeList_allowed = is_Elements_allowed

    def GetMacroInfo(self, macro_names):
        """GetMacroInfo(list<string> macro_names):

           Returns a list of string containing macro information.
           Each string is a JSON encoded.

           Params:
               - macro_name: a list of strings with the macro(s) name(s)
           Returns:
               - a list of string containing macro information.
        """
        macro_server = self.macro_server
        codec = CodecFactory().getCodec('json')
        ret = []

        for _, macro in list(macro_server.get_macros().items()):
            if macro.name in macro_names:
                ret.append(codec.encode(('', macro.serialize()))[1])

        return ret

    def ReloadMacro(self, macro_names):
        """ReloadMacro(list<string> macro_names):"""
        try:
            for macro_name in macro_names:
                self.macro_server.reload_macro(macro_name)
        except MacroServerException as mse:
            Except.throw_exception(mse.type, mse.msg, 'ReloadMacro')
        return ['OK']

    def ReloadMacroLib(self, lib_names):
        """ReloadMacroLib(sequence<string> lib_names):
        """
        try:
            for lib_name in lib_names:
                self.macro_server.reload_macro_lib(lib_name)
        except MacroServerException as mse:
            Except.throw_exception(mse.type, mse.msg, 'ReloadMacroLib')
        return ['OK']

    def GetMacroCode(self, argin):
        """GetMacroCode(<module name> [, <macro name>]) -> full filename, code, line_nb
        """
        ret = self.macro_server.get_or_create_macro_lib(*argin)
        return list(map(str, ret))

    def SetMacroCode(self, argin):
        lib_name, code = argin[:2]
        auto_reload = True
        if len(argin) > 2:
            auto_reload = argin[2].lower() in ('true', 'yes')
        self.macro_server.set_macro_lib(
            lib_name, code, auto_reload=auto_reload)

    #@DebugIt()
    def getEnvironment(self, cache=True):
        value = self.EnvironmentCache
        if cache and value is not None:
            return value
        env = self.macro_server.get_env()
        value = dict(new=env)
        value = CodecFactory().getCodec('pickle').encode(('', value))
        self.EnvironmentCache = value
        return value

    def read_Environment(self, attr):
        fmt, data = self.getEnvironment()
        attr.set_value(fmt, data)

    def write_Environment(self, attr):
        data = attr.get_write_value()
        data = CodecFactory().getCodec('pickle').decode(data)[1]
        self.macro_server.change_env(data)

    def is_Environment_allowed(self, req_type):
        return True


class MacroServerClass(SardanaDeviceClass):
    """MacroServer Tango class class"""

    #    Class Properties
    class_property_list = {
    }

    DefaultEnvBaseDir = "/tmp/tango"
    DefaultEnvRelDir = "%(ds_exec_name)s/%(ds_inst_name)s/macroserver.properties"

    DefaultLogReportFormat = '%(levelname)-8s %(asctime)s: %(message)s'

    #    Device Properties
    device_property_list = {
        'PoolNames':
            [DevVarStringArray,
             "Sardana device pool device names",
             []],
        'MacroPath':
            [DevVarStringArray,
             "list of directories to search for macros (path separators "
             "can be '\n' or character conventionally used by the OS to"
             "separate search path components, such as ':' for POSIX"
             "or ';' for Windows)",
             []],
        'RecorderPath':
            [DevVarStringArray,
             "list of directories to search for recorders (path separators "
             "can be '\n' or character conventionally used by the OS to"
             "separate search path components, such as ':' for POSIX"
             "or ';' for Windows)",
             []],
        'PythonPath':
            [DevVarStringArray,
             "list of directories to be appended to sys.path at startup (path "
             "separators can be '\n' or ':')",
             []],
        'MaxParallelMacros':
            [DevLong,
             "Maximum number of macros that can execute concurrently.",
             [10]],
        'EnvironmentDb':
            [DevString,
             "The environment database (usually a plain file).",
             os.path.join(DefaultEnvBaseDir, DefaultEnvRelDir)],
        'RConsolePort':
            [DevLong,
             "The rconsole port number",
             None],
        'LogReportFilename':
            [DevString,
             "Filename (absolute) which contains user log reports [default: "
             "None, meaning don't store log report messages]. The system will "
             "save old log files by appending extensions to the filename. The "
             "extensions are date-and-time based, using the strftime "
             "format %Y-%m-%d_%H-%M-%S or a leading portion thereof, "
             "depending on the rollover interval.",
             None],
        'LogReportFormat':
            [DevString,
             "Log report format [default: '%s']" % DefaultLogReportFormat,
             DefaultLogReportFormat],
        'LogstashHost':
            [DevString,
             "Hostname where Logstash runs. "
             "This property has been included in Sardana on a provisional "
             "basis. Backwards incompatible changes (up to and including "
             "its removal) may occur if deemed necessary by the "
             "core developers.",
             None],
        'LogstashPort':
            [DevLong,
             "Port on which Logstash will listen on events [default: 12345]. "
             "This property has been included in Sardana on a provisional "
             "basis. Backwards incompatible changes (up to and including "
             "its removal) may occur if deemed necessary by the "
             "core developers.",
             12345],
        'LogstashCacheDbPath':
            [DevString,
             "Path to the Logstash cache database [default: None]. "
             "It is advised not to use the database cache, as it may "
             "have negative effects on logging performance. See #895. "
             "This property has been included in Sardana on a provisional "
             "basis. Backwards incompatible changes (up to and including "
             "its removal) may occur if deemed necessary by the "
             "core developers.",
             None]
    }

    #    Command definitions
    cmd_list = {
        'GetMacroInfo':
            [[DevVarStringArray, "Macro(s) name(s)"],
             [DevVarStringArray, "Macro(s) description(s)"]],
        'ReloadMacro':
            [[DevVarStringArray, "Macro(s) name(s)"],
             [DevVarStringArray, "[OK] if successfull or a traceback "
                "if there was a error (one string with complete traceback of "
                "each error)"]],
        'ReloadMacroLib':
            [[DevVarStringArray, "MacroLib(s) name(s)"],
             [DevVarStringArray, "[OK] if successfull or a traceback "
                "if there was a error (one string with complete traceback of "
                "each error)"]],
        'GetMacroCode':
            [[DevVarStringArray, "<MacroLib name> [, <Macro name>]"],
             [DevVarStringArray, "result is a sequence of 3 strings:\n"
                "<full path and file name>, <code>, <line number>"]],
        'SetMacroCode':
            [[DevVarStringArray, "<MacroLib name>, <code> [, <Auto reload>=True]\n"
                "- if macro lib is a simple module name:\n"
                "  - if it exists, it is overwritten, otherwise a new python "
                "file is created in the directory of the first element in "
                "the MacroPath property"
                "- if macro lib is the full path name:\n"
                "  - if path is not in the MacroPath, an exception is thrown"
                "  - if file exists it is overwritten otherwise a new file "
                "is created"],
             [DevVoid, ""]],
    }

    #    Attribute definitions
    attr_list = {
        'DoorList': [[DevString,  SPECTRUM, READ, 256]],
        'MacroList': [[DevString,  SPECTRUM, READ, 4096]],
        'MacroLibList': [[DevString,  SPECTRUM, READ, 1024]],
        'TypeList': [[DevString,  SPECTRUM, READ, 256]],
        'Elements': [[DevEncoded, SCALAR,   READ],
                     {'label': "Elements",
                      'description': "the list of all elements "
                      "(a JSON encoded dict)", }],
        'Environment': [[DevEncoded, SCALAR, READ_WRITE],
                        {'label': 'Environment',
                         'description': "The macro server environment "
                         "(a JSON encoded dict)", }],
    }



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of a discrete pseudo motor controller
for the Sardana Device Pool"""

__all__ = ["DiscretePseudoMotorController"]

__docformat__ = 'restructuredtext'

import json

from sardana import DataAccess
from sardana.pool.controller import PseudoMotorController
from sardana.pool.controller import Type, Access, Description


class DiscretePseudoMotorController(PseudoMotorController):
    """
    A discrete pseudo motor controller which converts physical motor
    positions to discrete values"""

    gender = "DiscretePseudoMotorController"
    model = "PseudoMotor"
    organization = "Sardana team"
    image = ""

    pseudo_motor_roles = ("DiscreteMoveable",)
    motor_roles = ("ContinuousMoveable",)

    axis_attributes = {'Configuration':
                       # type hackish until encoded attributes supported
                       {Type: str,
                        Description: 'String dictionary mapping the labels'
                                     ' and discrete positions',
                        Access: DataAccess.ReadWrite}
                       }

    def __init__(self, inst, props, *args, **kwargs):
        PseudoMotorController.__init__(self, inst, props, *args, **kwargs)
        self._calibration = []
        self._positions = []
        self._labels = []
        self._configuration = None
        self._calibration_cfg = None
        self._positions_cfg = None
        self._labels_cfg = None

    def GetAxisAttributes(self, axis):
        axis_attrs = PseudoMotorController.GetAxisAttributes(self, axis)
        axis_attrs = dict(axis_attrs)
        axis_attrs['Position']['type'] = float
        return axis_attrs

    def CalcPseudo(self, axis, physical_pos, curr_pseudo_pos):
        positions = self._positions_cfg
        calibration = self._calibration_cfg
        labels = self._labels_cfg

        llabels = len(labels)
        lcalibration = len(calibration)

        value = physical_pos[0]
        # case 0: nothing to translate, only round about integer the attribute
        # value
        if llabels == 0:
            return int(value)
        # case 1: only uses the labels. Available positions in POSITIONS
        elif lcalibration == 0:
            value = int(value)
            try:
                positions.index(value)
            except Exception:
                raise Exception("Invalid position.")
            else:
                return value
        # case 1+fussy: the physical position must be in one of the defined
        # ranges, and the DiscretePseudoMotor position is defined in labels
        elif llabels == lcalibration:
            for fussyPos in calibration:
                if value >= fussyPos[0] and value <= fussyPos[2]:
                    return positions[calibration.index(fussyPos)]
            # if the loop ends, current value is not in the fussy areas.
            raise Exception("Invalid position.")
        else:
            raise Exception("Bad configuration on axis attributes.")

    def CalcPhysical(self, axis, pseudo_pos, curr_physical_pos):
        positions = self._positions_cfg
        calibration = self._calibration_cfg
        labels = self._labels_cfg

        # If Labels is well defined, the write value must be one this struct
        llabels = len(labels)
        lcalibration = len(calibration)
        value = pseudo_pos[0]

        # case 0: nothing to translate, what is written goes to the attribute
        if llabels == 0:
            return value
        # case 1: only uses the labels. Available positions in POSITIONS
        elif lcalibration == 0:
            self._log.debug("Value = %s", value)
            try:
                positions.index(value)
            except Exception:
                raise Exception("Invalid position.")
            return value
        # case 1+fussy: the write to the to the DiscretePseudoMotorController
        # is translated to the central position of the calibration.
        elif llabels == lcalibration:
            self._log.debug("Value = %s", value)
            try:
                destination = positions.index(value)
            except Exception:
                raise Exception("Invalid position.")
            self._log.debug("destination = %s", destination)
            calibrated_position = calibration[
                destination][1]  # central element
            self._log.debug("calibrated_position = %s", calibrated_position)
            return calibrated_position

    def getConfiguration(self, axis):
        return json.dumps(self._configuration)

    def setConfiguration(self, axis, value):
        try:
            mapping = json.loads(value)
            labels = []
            positions = []
            calibration = []
            for k, v in list(mapping.items()):
                labels.append(k)
                pos = int(v['pos'])
                if pos in positions:
                    msg = 'position {0} is already used'.format(pos)
                    raise ValueError(msg)
                positions.append(pos)
                if all([x in list(v.keys()) for x in ['min', 'set', 'max']]):
                    calibration.append([v['min'], v['set'], v['max']])
            self._labels_cfg = labels
            self._positions_cfg = positions
            self._calibration_cfg = calibration
            self._configuration = json.loads(value)
        except Exception as e:
            msg = "invalid configuration: {0}".format(e)
            raise Exception(msg)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
motor.py:
"""


from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.util.ui import UILoadable


def showDialogConfigureMotor(parent):
    Dialog = Qt.QDialog(parent)
    Dialog.resize((Qt.QSize(Qt.QRect(0, 0, 310, 309).size()
                            ).expandedTo(Dialog.minimumSizeHint())))
    motorV2 = TaurusMotorV2(Dialog)
    motorV2.setModel(parent.model)
    motorV2.setGeometry(Qt.QRect(10, 10, 291, 291))
    Dialog.show()


@UILoadable(with_ui='ui')
class TaurusMotorH(Qt.QWidget, TaurusBaseWidget):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TaurusBaseWidget, str(
            self.objectName()), designMode=designMode)
        self.loadUi()
        self.ui.config.clicked.connect(self.configureMotor)

    def sizeHint(self):
        return Qt.QSize(330, 50)

    def configureMotor(self):
        showDialogConfigureMotor(self.ui.TaurusGroupBox)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None
#        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
#        ret['module'] = 'taurus.qt.qtgui.extra_pool'
#        ret['group'] = 'Taurus Sardana'
#        ret['icon'] = ':/designer/extra_pool.png'
#        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    @Qt.pyqtSlot()
    def getModel(self):
        return self.ui.TaurusGroupBox.getModel()

    @Qt.pyqtSlot("QString")
    def setModel(self, model):
        self.ui.TaurusGroupBox.setModel(model)

    @Qt.pyqtSlot()
    def resetModel(self):
        self.ui.TaurusGroupBox.resetModel()

    @Qt.pyqtSlot()
    def getShowText(self):
        return self.ui.TaurusGroupBox.getShowText()

    @Qt.pyqtSlot(bool)
    def setShowText(self, showText):
        self.ui.TaurusGroupBox.setShowText(showText)

    @Qt.pyqtSlot()
    def resetShowText(self):
        self.ui.TaurusGroupBox.resetShowText()

    model = Qt.pyqtProperty("QString", getModel, setModel, resetModel)
    showText = Qt.pyqtProperty("bool", getShowText, setShowText, resetShowText)


@UILoadable(with_ui='ui')
class TaurusMotorH2(Qt.QWidget, TaurusBaseWidget):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TaurusBaseWidget, str(
            self.objectName()), designMode=designMode)
        self.loadUi()
        self.ui.config.clicked.connect(self.configureMotor)

    def sizeHint(self):
        return Qt.QSize(215, 85)

    def configureMotor(self):
        showDialogConfigureMotor(self.ui.TaurusGroupBox)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None
#        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
#        ret['module'] = 'taurus.qt.qtgui.extra_pool'
#        ret['group'] = 'Taurus Sardana'
#        ret['icon'] = ':/designer/extra_pool.png'
#        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    @Qt.pyqtSlot()
    def getModel(self):
        return self.ui.TaurusGroupBox.getModel()

    @Qt.pyqtSlot("QString")
    def setModel(self, model):
        self.ui.TaurusGroupBox.setModel(model)

    @Qt.pyqtSlot()
    def resetModel(self):
        self.ui.TaurusGroupBox.resetModel()

    @Qt.pyqtSlot()
    def getShowText(self):
        return self.ui.TaurusGroupBox.getShowText()

    @Qt.pyqtSlot(bool)
    def setShowText(self, showText):
        self.ui.TaurusGroupBox.setShowText(showText)

    @Qt.pyqtSlot()
    def resetShowText(self):
        self.ui.TaurusGroupBox.resetShowText()

    model = Qt.pyqtProperty("QString", getModel, setModel, resetModel)
    showText = Qt.pyqtProperty("bool", getShowText, setShowText, resetShowText)


@UILoadable(with_ui='ui')
class TaurusMotorV(Qt.QWidget, TaurusBaseWidget):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TaurusBaseWidget, str(
            self.objectName()), designMode=designMode)
        self.loadUi()
        self.ui.config.clicked.connect(self.configureMotor)

    def sizeHint(self):
        return Qt.QSize(120, 145)

    def configureMotor(self):
        showDialogConfigureMotor(self.ui.TaurusGroupBox)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None
#        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
#        ret['module'] = 'taurus.qt.qtgui.extra_pool'
#        ret['group'] = 'Taurus Sardana'
#        ret['icon'] = ':/designer/extra_pool.png'
#        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    @Qt.pyqtSlot()
    def getModel(self):
        return self.ui.TaurusGroupBox.getModel()

    @Qt.pyqtSlot("QString")
    def setModel(self, model):
        self.ui.TaurusGroupBox.setModel(model)

    @Qt.pyqtSlot()
    def resetModel(self):
        self.ui.TaurusGroupBox.resetModel()

    @Qt.pyqtSlot()
    def getShowText(self):
        return self.ui.TaurusGroupBox.getShowText()

    @Qt.pyqtSlot(bool)
    def setShowText(self, showText):
        self.ui.TaurusGroupBox.setShowText(showText)

    @Qt.pyqtSlot()
    def resetShowText(self):
        self.ui.TaurusGroupBox.resetShowText()

    model = Qt.pyqtProperty("QString", getModel, setModel, resetModel)
    showText = Qt.pyqtProperty("bool", getShowText, setShowText, resetShowText)


@UILoadable(with_ui='ui')
class TaurusMotorV2(Qt.QWidget, TaurusBaseWidget):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TaurusBaseWidget, str(
            self.objectName()), designMode=designMode)
        self.loadUi()

    def sizeHint(self):
        return Qt.QSize(300, 275)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None
#        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
#        ret['module'] = 'taurus.qt.qtgui.extra_pool'
#        ret['group'] = 'Taurus Sardana'
#        ret['icon'] = ':/designer/extra_pool.png'
#        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    @Qt.pyqtSlot()
    def getModel(self):
        return self.ui.TaurusGroupBox.getModel()

    @Qt.pyqtSlot("QString")
    def setModel(self, model):
        self.ui.TaurusGroupBox.setModel(model)

    @Qt.pyqtSlot()
    def resetModel(self):
        self.ui.TaurusGroupBox.resetModel()

    @Qt.pyqtSlot()
    def getShowText(self):
        return self.ui.TaurusGroupBox.getShowText()

    @Qt.pyqtSlot(bool)
    def setShowText(self, showText):
        self.ui.TaurusGroupBox.setShowText(showText)

    @Qt.pyqtSlot()
    def resetShowText(self):
        self.ui.TaurusGroupBox.resetShowText()

    model = Qt.pyqtProperty("QString", getModel, setModel, resetModel)
    showText = Qt.pyqtProperty("bool", getShowText, setShowText, resetShowText)


if __name__ == "__main__":

    import sys
    app = Qt.QApplication(sys.argv)

    form = TaurusMotorH()
    form.setModel(sys.argv[1])

    form.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Base classes for the controller tests"""

import PyTango
import unittest
from taurus.core.tango.starter import ProcessStarter
from sardana import sardanacustomsettings
from sardana.tango.core.util import (get_free_server, get_free_device,
                                     get_free_alias)
from taurus.core.util import whichexecutable

__all__ = ['BasePoolTestCase', 'ControllerLoadsTestCase',
           'ControllerCreationTestCase', 'ElementCreationTestCase']


class BasePoolTestCase(object):
    """Abstract class for pool DS testing.
    """
    pool_ds_name = getattr(sardanacustomsettings, 'UNITTEST_POOL_DS_NAME')
    pool_name = getattr(sardanacustomsettings, 'UNITTEST_POOL_NAME')

    def setUp(self, properties=None):
        """Start Pool DS and register extensions.
        """
        db = PyTango.Database()
        # Discover the Pool launcher script
        poolExec = whichexecutable.whichfile("Pool")
        # register Pool server
        pool_ds_name = "Pool/" + self.pool_ds_name
        pool_free_ds_name = get_free_server(PyTango.Database(),
                                            pool_ds_name)
        self._starter = ProcessStarter(poolExec, pool_free_ds_name)
        # register Pool device
        dev_name_parts = self.pool_name.split('/')
        prefix = '/'.join(dev_name_parts[0:2])
        start_from = int(dev_name_parts[2])
        self.pool_name = get_free_device(db, prefix, start_from)
        self._starter.addNewDevice(self.pool_name, klass='Pool')
        # Add properties
        if properties is not None:
            for key, values in list(properties.items()):
                db.put_device_property(self.pool_name,
                                       {key: values})
        # start Pool server
        self._starter.startDs(wait_seconds=20)
        # register extensions so the test methods can use them
        self.pool = PyTango.DeviceProxy(self.pool_name)

    def tearDown(self):
        """Remove the Pool instance.
        """
        self._starter.cleanDb(force=True)
        self._starter = None
        self.pool = None
        self.pool_name = None


# TODO: Currently test inputs are implemented as class members, it would be
# more aesthetic to implement them as decorators.
class ControllerLoadsTestCase(BasePoolTestCase):
    """Class for loading an arbitrary Sardana controller library and class.
    """
    controller_classes = []

    def test_controller_loads(self):
        """Test that the controller library and class can be loaded.
        """
        libraries = \
            list(self.pool.getElementsOfType('ControllerLibrary').values())
        libraries_names = [lib.getName() for lib in libraries]
        classes = \
            list(self.pool.getElementsOfType('ControllerClass').values())
        classes_names = [cls.getName() for cls in classes]

        for test_lib, test_classes in list(self.controller_classes.items()):
            msg = 'ControllerLibrary %s was not correctly loaded.' % test_lib
            self.assertIn(test_lib, libraries_names, msg)
            msg = 'ControllerClass %s was not correctly loaded.'
            for test_class in test_classes:
                self.assertIn(test_class, classes_names, msg % test_class)


# TODO: Currently test inputs are implemented as class members, it would be
# more aesthetic to implement them as decorators.
class ControllerCreationTestCase(BasePoolTestCase):
    """Class for creating a controller and testing the correct creation.
    """
    controller_infos = []

    def test_controller_creation(self):
        """Test that the controller has been created with the correct name.
        """
        for cls, name, props in self.controller_infos:
            ctrl = self.pool.createController(cls, name, *props)
            msg = 'Controller %s was not correctly created.' % name
            self.assertEqual(ctrl.getName(), name, msg)
            ctrl = self.pool.deleteElement(ctrl.getName())


# TODO: Currently test inputs are implemented as class members, it would be
# more aesthetic to implement them as decorators.
class ElementCreationTestCase(BasePoolTestCase):
    """Class used for creating a Sardana controller and Sardana elements.
    """
    controller_infos = []
    NAME = 0
    AXIS = 1

    def test_element_creation(self):
        """Test that controller and elements have been correctly created.
        """
        for cls, name, props, elements in self.controller_infos:
            ctrl = self.pool.createController(cls, name, *props)
            msg = 'Controller %s was not correctly created.' % name
            self.assertEqual(ctrl.getName(), name, msg)
            for element_info in elements:
                test_name = element_info[self.NAME]
                test_axis = element_info[self.AXIS]
                elem = self.pool.createElement(test_name, ctrl, test_axis)
                msg = 'Element %s was not correctly created.' % test_name
                self.assertIsNotNone(elem, msg)
                name = elem.getName()
                msg = 'Element name: %s does not correspond to: %s.' % \
                      (name, test_name)
                self.assertEqual(name, test_name, msg)
                elem = self.pool.deleteElement(test_name)
                msg = 'Element %s was not correctly deleted.' % test_name
                self.assertIsNotNone(elem, msg)
            ctrl = self.pool.deleteElement(ctrl.getName())


if __name__ == '__main__':

    class BuiltinControllerLoadsTest(ControllerLoadsTestCase,
                                     unittest.TestCase):

        controller_classes = {
            'DummyMotorController': ('DummyMotorController',)
        }

    class BuiltinControllerCreationTest(ControllerCreationTestCase,
                                        unittest.TestCase):

        controller_infos = [('DummyMotorController', 'unittest', ())
                            ]

    class BuiltinElementCreationTest(ElementCreationTestCase,
                                     unittest.TestCase):
        alias = get_free_alias(PyTango.Database(), "mot_test")
        controller_infos = [('DummyMotorController',
                             'unittest',
                             (),
                             [(alias, 1)])
                            ]

    suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        BuiltinElementCreationTest)
    unittest.TextTestRunner(descriptions=True, verbosity=2).run(suite)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for"""

__all__ = ["PoolMeasurementGroup", "MeasurementConfiguration",
           "ControllerConfiguration", "ChannelConfiguration",
           "SynchronizerConfiguration", "build_measurement_configuration"]

__docformat__ = 'restructuredtext'

import threading
import weakref


from taurus.core.tango.tangovalidator import TangoAttributeNameValidator

from sardana import State, ElementType, TYPE_EXP_CHANNEL_ELEMENTS
from sardana.sardanaevent import EventType
from sardana.pool.pooldefs import AcqMode, SynchParam, AcqSynch, \
    SynchDomain, AcqSynchType

from sardana.pool.poolgroupelement import PoolGroupElement
from sardana.pool.poolacquisition import PoolAcquisition
from sardana.pool.poolsynchronization import SynchDescription
from sardana.pool.poolexternal import PoolExternalObject

from sardana.taurus.core.tango.sardana import PlotType, Normalization


# ----------------------------------------------
# Measurement Group Configuration information
# ----------------------------------------------
# dict <str, obj> with (at least) keys:
#    - 'timer' : the timer channel name / timer channel id
#    - 'monitor' : the monitor channel name / monitor channel id
#    - 'controllers' : dict<Controller, dict> where:
#        - key: ctrl
#        - value: dict<str, dict> with (at least) keys:
#                - 'timer' : the timer channel name / timer channel id
#                - 'monitor' : the monitor channel name / monitor channel id
#                - 'synchronization' : 'Gate'/'Software'
#                - 'channels' where value is a dict<str, obj> with (at least)
#                   keys:
#                    - 'id' : the channel name ( channel id )
#                    optional keys:
#                    - 'enabled' : True/False (default is True)
#                    any hints:
#                    - 'output' : True/False (default is True)
#                    - 'plot_type' : 'No'/'1D'/'2D' (default is 'No')
#                    - 'plot_axes' : list<str> 'where str is channel
#                                    name/'step#/'index#' (default is [])
#                    - 'label' : prefered label (default is channel name)
#                    - 'scale' : <float, float> with min/max (defaults to
#                                channel range if it is defined
#                    - 'plot_color' : int representing RGB
#    optional keys:
#    - 'label' : measurement group label (defaults to measurement group name)
#    - 'description' : measurement group description

# <MeasurementGroupConfiguration>
#  <timer>UxTimer</timer>
#  <monitor>CT1</monitor>
# </MeasurementGroupConfiguration>

# Example: 2 NI cards, where channel 1 of card 1 is wired to channel 1 of
# card 2 at configuration time we should set:

# ni0ctrl.setCtrlPar(0, 'synchronization', AcqSynch.SoftwareTrigger)
# ni0ctrl.setCtrlPar(0, 'timer', 1) # channel 1 is the timer
# ni0ctrl.setCtrlPar(0, 'monitor', 4) # channel 4 is the monitor
# ni1ctrl.setCtrlPar(0, 'synchronization', AcqSynch.HardwareTrigger)
# ni1ctrl.setCtrlPar(0, 'master', 0)

# when we count for 1.5 seconds:
# ni1ctrl.Load(1.5)
# ni0ctrl.Load(1.5)
# ni1ctrl.Start()
# ni0ctrl.Start()

"""

"""


def _to_fqdn(name, logger=None):
    full_name = name
    # try to use Taurus 4 to retrieve FQDN
    try:
        from taurus.core.tango.tangovalidator import TangoDeviceNameValidator
        full_name, _, _ = TangoDeviceNameValidator().getNames(name)
    # if Taurus3 in use just continue
    except ImportError:
        pass
    if full_name is None:
        full_name = name
    if full_name != name and logger:
        msg = ("PQDN full name is deprecated in favor of FQDN full name."
               " Re-apply configuration in order to upgrade.")
        logger.warning(msg)
    return full_name


class ConfigurationItem(object):
    """Container of configuration attributes related to a given element.

    Wrap an element to pretend its API.
    Manage the element's configuration.
    Hold an information whether the element is enabled.
    By default it is enabled.

    .. note::
        The ConfigurationItem class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, element, attrs=None):
        """Construct a wrapper around the element

        :param element: element to wrap
        :type element: obj
        :param: attrs: configuration attributes and their values
        :type attrs: dict
        """
        self._element = weakref.ref(element)
        self.enabled = True

        if attrs is not None:
            self.__dict__.update(attrs)

    def __getattr__(self, item):
        return getattr(self.element, item)

    def get_element(self):
        """Returns the element associated with this item"""
        return self._element()

    def set_element(self, element):
        """Sets the element for this item"""
        self._element = weakref.ref(element)

    element = property(get_element)


class ControllerConfiguration(ConfigurationItem):
    """Container of configuration attributes related to a given controller.

    Inherit behavior from
    :class:`~sardana.pool.poolmeasurementgroup.ConfigurationItem`
    and additionally hold information about its enabled/disabled channels.
    By default it is disabled.

    .. note::
        The ControllerConfiguration class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, element, attrs=None):
        ConfigurationItem.__init__(self, element, attrs)
        self.enabled = False
        self._channels = []
        self._channels_enabled = []
        self._channels_disabled = []

    def add_channel(self, channel_item):
        """Aggregate a channel configuration item."""
        self._channels.append(channel_item)
        if channel_item.enabled:
            self.enabled = True
            if self._channels_enabled is None:
                self._channels_enabled = []
            self._channels_enabled.append(channel_item)
        else:
            if self._channels_disabled is None:
                self._channels_disabled = []
            self._channels_disabled.append(channel_item)

    def remove_channel(self, channel_item):
        """Remove a channel configuration item."""
        self._channels.remove(channel_item)
        if channel_item.enabled:
            self._channels_enabled.remove(channel_item)
            if len(self._channels_enabled) == 0:
                self.enabled = False
        else:
            self._channels_disabled.remove(channel_item)

    def update_state(self):
        """Update internal state based on the aggregated channels."""
        self.enabled = False
        self._channels_enabled = []
        self._channels_disabled = []
        for channel_item in self._channels:
            if channel_item.enabled:
                self.enabled = True
                self._channels_enabled.append(channel_item)
            else:
                self._channels_disabled.append(channel_item)

    def get_channels(self, enabled=None):
        """Return aggregated channels.

        :param enabled: which channels to return
         - True - only enabled
         - False - only disabled
         - None - all

        :type enabled: bool or None
        """
        if enabled is None:
            return list(self._channels)
        elif enabled:
            return list(self._channels_enabled)
        else:
            return list(self._channels_disabled)

    def validate(self):
        pass


class TimerableControllerConfiguration(ControllerConfiguration):
    """Container of configuration attributes related to a given
    timerable controller.

    Inherit behavior from
    :class:`~sardana.pool.poolmeasurementgroup.ControllerConfiguration`
    and additionally validate *timer* and *monitor* configuration.

    .. note::
        The TimerableControllerConfiguration class has been included in
        Sardana on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """

    def update_timer(self):
        self._update_master("timer")

    def update_monitor(self):
        self._update_master("monitor")

    def _update_master(self, role):
        master = getattr(self, role, None)
        if master is None:
            idx = float("+inf")
            for channel in self._channels_enabled:
                if channel.index > idx:
                    continue
                master = channel
                idx = channel.index
        else:
            found = False
            for channel in self._channels:
                if channel.full_name == master:
                    master = channel
                    found = True
                    break
            if not found:
                master = None
        setattr(self, role, master)

    def validate(self):
        # validate if the timer and monitor are disabled if the
        # controller is enabled
        if self.enabled \
                and not self.timer.enabled \
                and not self.monitor.enabled:
            err_msg = 'channel {0} used as timer and channel ' \
                      '{1} used as monitor are disabled. One of them ' \
                      'must be enabled.'.format(self.timer.name,
                                                self.monitor.name)
            raise ValueError(err_msg)


class ExternalControllerConfiguration(ControllerConfiguration):
    """Container of configuration attributes related to a given
    external controller.

    Inherit behavior from
    :class:`~sardana.pool.poolmeasurementgroup.ControllerConfiguration`.

    .. note::
        The ExternalControllerConfiguration class has been included in
        Sardana on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, element, attrs=None):
        ControllerConfiguration.__init__(self, self, attrs)
        self.full_name = element


class ChannelConfiguration(ConfigurationItem):
    """Container of configuration attributes related to a given
    experimental channel.

    Inherit behavior from
    :class:`~sardana.pool.poolmeasurementgroup.ConfigurationItem`.

    .. note::
        The ChannelConfiguration class has been included in
        Sardana on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """


class SynchronizerConfiguration(ConfigurationItem):
    """Container of configuration attributes related to a given
    synchronizer element.

    Inherit behavior from
    :class:`~sardana.pool.poolmeasurementgroup.ConfigurationItem`.
    By default it is disabled.

    .. note::
        The ChannelConfiguration class has been included in
        Sardana on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, element, attrs=None):
        ConfigurationItem.__init__(self, element, attrs)
        self.enabled = False


def build_measurement_configuration(user_elements):
    """Create a minimal measurement configuration data structure from the
    user_elements list.

    .. highlight:: none

    Minimal configuration data structure::

        dict <str, dict> with keys:
        - 'controllers' : where value is a dict<str, dict> where:
            - key: controller's full name
            - value: dict<str, dict> with keys:
                - 'channels' where value is a dict<str, obj> where:
                    - key: channel's full name
                    - value: dict<str, obj> with keys:
                        - 'index' : where value is the channel's index <int>

    .. highlight:: default

    .. note::
        The build_measurement_configuration function has been included in
        Sardana on a provisional basis. Backwards incompatible changes
        (up to and including removal of the function) may occur if
        deemed necessary by the core developers.
    """
    user_config = {}
    external_user_elements = []
    user_config["controllers"] = controllers = {}

    for index, element in enumerate(user_elements):
        elem_type = element.get_type()
        if elem_type == ElementType.External:
            external_user_elements.append((index, element))
            continue

        ctrl = element.controller
        ctrl_data = controllers.get(ctrl.full_name)

        if ctrl_data is None:
            controllers[ctrl.full_name] = ctrl_data = {}
            ctrl_data['channels'] = channels = {}
        else:
            channels = ctrl_data['channels']
        channels[element.full_name] = channel_data = {}
        channel_data['index'] = index

    if len(external_user_elements) > 0:
        controllers['__tango__'] = ctrl_data = {}
        ctrl_data['channels'] = channels = {}
        for index, element in external_user_elements:
            channels[element.full_name] = channel_data = {}
            channel_data['index'] = index
    return user_config


class MeasurementConfiguration(object):
    """Configuration of a measurement.

    Accepts import and export from/to a serializable data structure (based on
    dictionaries/lists and strings).
    Provides getter methods that facilitate extracting of information e.g.
    controllers of different types, master timers/monitors, etc.

    .. note::
        The build_measurement_configuration function has been included in
        Sardana on a provisional basis. Backwards incompatible changes
        (up to and including removal of the function) may occur if
        deemed necessary by the core developers.
    """

    DFT_DESC = 'General purpose measurement configuration'

    def __init__(self, parent=None):
        """Initialize measurement configuration object

        :param parent: (optional) object that this measurement configuration
        refers to (usually
         :class:`~sardana.pool.poolmeasurementgroup.PoolMeasurementGroup)`
        """
        self._parent = None
        if parent is not None:
            self._parent = weakref.proxy(parent)
        self._config = None
        # Structure to store the controllers and their channels
        self._timerable_ctrls = {}
        self._zerod_ctrls = []
        self._synch_ctrls = {}
        self._other_ctrls = []
        self._master_timer_sw = None
        self._master_monitor_sw = None
        self._master_timer_sw_start = None
        self._master_monitor_sw_start = None
        self._label = None
        self._description = None
        self._user_config = {}
        self._channel_acq_synch = {}
        self._ctrl_acq_synch = {}
        self.changed = False
        # provide back. compatibility for value_ref_{enabled,pattern}
        # config parameters created with Sardana < 3.
        self._value_ref_compat = False

    def get_acq_synch_by_channel(self, channel):
        """Return acquisition synchronization configured for this element.

        :param channel: channel to look for its acquisition synchronization
        :type channel: :class:`~sardana.pool.poolbasechannel.PoolBaseChannel`
         or :class:`~sardana.pool.poolmeasurementgroup.ChannelConfiguration`
        :return: acquisition synchronization
        :rtype: :obj:`~sardana.pool.pooldefs.AcqSynch`
        """
        if isinstance(channel, ChannelConfiguration):
            channel = channel.element
        return self._channel_acq_synch[channel]

    def get_acq_synch_by_controller(self, controller):
        """Return acquisition synchronization configured for this controller.

        :param controller: controller to look for its acquisition
         synchronization
        :type controller: :class:`~sardana.pool.poolcontroller.PoolController`
         or :class:`~sardana.pool.poolmeasurementgroup.ControllerConfiguration`
        :return: acquisition synchronization
        :rtype: :obj:`~sardana.pool.pooldefs.AcqSynch`
        """
        if isinstance(controller, ConfigurationItem):
            controller = controller.element
        return self._ctrl_acq_synch[controller]

    def _filter_ctrls(self, ctrls, enabled):
        if enabled is None:
            return ctrls

        filtered_ctrls = []
        for ctrl in ctrls:
            if ctrl.enabled == enabled:
                filtered_ctrls.append(ctrl)
        return filtered_ctrls

    def get_timerable_ctrls(self, acq_synch=None, enabled=None):
        """Return timerable controllers.

        Allow to filter controllers based on acquisition synchronization or
        whether these are enabled/disabled.

        :param acq_synch: (optional) filter controller based on acquisition
         synchronization
        :type acq_synch: :class:`~sardana.pool.pooldefs.AcqSynch`
        :param enabled: (optional) filter controllers whether these are
         enabled/disabled:

         - :obj:`True` - enabled only
         - :obj:`False` - disabled only
         - :obj:`None` - all

        :type enabled: :obj:`bool` or :obj:`None`
        :return: timerable controllers that fulfils the filtering criteria
        :rtype: list<:class:`~sardana.pool.poolmeasurementgroup.ControllerConfiguration`>  # noqa
        """
        timerable_ctrls = []
        if acq_synch is None:
            for ctrls in list(self._timerable_ctrls.values()):
                timerable_ctrls += ctrls
        elif isinstance(acq_synch, list):
            acq_synch_list = acq_synch
            for acq_synch in acq_synch_list:
                timerable_ctrls += self._timerable_ctrls[acq_synch]
        else:
            timerable_ctrls = list(self._timerable_ctrls[acq_synch])

        return self._filter_ctrls(timerable_ctrls, enabled)

    def get_zerod_ctrls(self, enabled=None):
        """Return 0D controllers.

        Allow to filter controllers whether these are enabled/disabled.

        :param enabled: (optional) filter controllers whether these are
         enabled/disabled:

         - :obj:`True` - enabled only
         - :obj:`False` - disabled only
         - :obj:`None` - all

        :type enabled: :obj:`bool` or :obj:`None`
        :return: 0D controllers that fulfils the filtering criteria
        :rtype: list<:class:`~sardana.pool.poolmeasurementgroup.ControllerConfiguration`>  # noqa
        """
        return self._filter_ctrls(self._zerod_ctrls, enabled)

    def get_synch_ctrls(self, enabled=None):
        """Return synchronizer (currently only trigger/gate) controllers.

        Allow to filter controllers whether these are enabled/disabled.

        :param enabled: (optional) filter controllers whether these are
         enabled/disabled:

         - :obj:`True` - enabled only
         - :obj:`False` - disabled only
         - :obj:`None` - all

        :type enabled: :obj:`bool` or :obj:`None`
        :return: synchronizer controllers that fulfils the filtering criteria
        :rtype: list<:class:`~sardana.pool.poolmeasurementgroup.ControllerConfiguration`>  # noqa
        """
        return self._filter_ctrls(self._synch_ctrls, enabled)

    def get_master_timer_software(self):
        """Return master timer in software acquisition.

        :return: master timer in software acquisition
        :rtype: :class:`~sardana.pool.poolmeasurementgroup.ChannelConfiguration`  # noqa
        """
        return self._master_timer_sw

    def get_master_monitor_software(self):
        """Return master monitor in software acquisition.

        :return: master monitor in software acquisition
        :rtype: :class:`~sardana.pool.poolmeasurementgroup.ChannelConfiguration`  # noqa
        """
        return self._master_monitor_sw

    def get_master_timer_software_start(self):
        """Return master timer in software start acquisition.

        :return: master timer in software start acquisition
        :rtype: :class:`~sardana.pool.poolmeasurementgroup.ChannelConfiguration`  # noqa
        """
        return self._master_monitor_sw_start

    def get_master_monitor_software_start(self):
        """Return master monitor in software start acquisition.

        :return: master monitor in software start acquisition
        :rtype: :class:`~sardana.pool.poolmeasurementgroup.ChannelConfiguration`  # noqa
        """
        return self._master_timer_sw_start

    def get_configuration_for_user(self):
        """Return measurement configuration serializable data structure."""
        return self._user_config

    def set_configuration_from_user(self, cfg):
        """Set measurement configuration from serializable data structure.

        Setting of the configuration includes the validation process. Setting
        of invalid configuration raises an exception hence it is not necessary
        that the client application does the validation.

        The configuration parameters for given channels/controllers may differ
        depending on their types e.g. 0D channel does not support timer
        parameter while C/T does.

        .. todo::
            Raise exceptions when setting _Synchronization_ parameter for
            external channels, 0D and PSeudoCounters.
        """

        pool = self._parent.pool

        label = cfg.get('label', self._parent.name)
        description = cfg.get('description', self.DFT_DESC)

        timerable_ctrls = {AcqSynch.HardwareGate: [],
                           AcqSynch.HardwareStart: [],
                           AcqSynch.HardwareTrigger: [],
                           AcqSynch.SoftwareStart: [],
                           AcqSynch.SoftwareTrigger: [],
                           AcqSynch.SoftwareGate: []}
        zerod_ctrls = []
        synch_ctrls = []
        other_ctrls = []
        master_timer_sw = None
        master_monitor_sw = None
        master_timer_sw_start = None
        master_monitor_sw_start = None
        master_timer_idx_sw = float("+inf")
        master_monitor_idx_sw = float("+inf")
        master_timer_idx_sw_start = float("+inf")
        master_monitor_idx_sw_start = float("+inf")
        user_elem_ids = {}
        channel_acq_synch = {}
        ctrl_acq_synch = {}
        user_config = {}

        user_config['controllers'] = {}
        user_config['label'] = label
        user_config['description'] = description

        for ctrl_name, ctrl_data in list(cfg['controllers'].items()):
            # backwards compatibility for measurement groups created before
            # implementing feature-372:
            # https://sourceforge.net/p/sardana/tickets/372/
            # WARNING: this is one direction backwards compatibility - it just
            # reads channels from the units, but does not write channels to the
            # units back
            if 'units' in ctrl_data:
                ctrl_data = ctrl_data['units']['0']
            # discard controllers which don't have items (garbage)
            ch_count = len(ctrl_data['channels'])
            if ch_count == 0:
                continue

            external = ctrl_name in ['__tango__']

            if external:
                ctrl = ctrl_name
            else:
                ctrl = pool.get_element_by_full_name(ctrl_name)
                assert ctrl.get_type() == ElementType.Controller

            user_config['controllers'][ctrl_name] = user_config_ctrl = {}
            ctrl_conf = {}
            conf_synch = None

            # The external controllers should not have synchronizer

            if external:
                if 'synchronizer' in ctrl_data:
                    raise ValueError('External controller does not allow '
                                     'to have synchronizer')
                if 'monitor' in ctrl_data:
                    raise ValueError('External controller does not allow '
                                     'to have monitor')
                if 'timer' in ctrl_data:
                    raise ValueError('External controller does not allow '
                                     'to have timer')
            else:
                synchronizer = ctrl_data.get('synchronizer', 'software')
                if synchronizer is None or synchronizer == 'software':
                    ctrl_conf['synchronizer'] = 'software'
                    user_config_ctrl['synchronizer'] = 'software'
                else:
                    user_config_ctrl['synchronizer'] = synchronizer
                    pool_synch = pool.get_element_by_full_name(synchronizer)
                    pool_synch_ctrl = pool_synch.controller
                    conf_synch_ctrl = None
                    conf_synch = None
                    for conf_ctrl_created in synch_ctrls:
                        if pool_synch_ctrl == conf_ctrl_created.element:
                            conf_synch_ctrl = conf_ctrl_created
                            for conf_synch_created in \
                                    conf_ctrl_created.get_channels():
                                if pool_synch == conf_synch_created.element:
                                    conf_synch = conf_synch_created
                                    break
                            break

                    if conf_synch_ctrl is None:
                        conf_synch_ctrl = \
                            ControllerConfiguration(pool_synch_ctrl)
                        synch_ctrls.append(conf_synch_ctrl)

                    if conf_synch is None:
                        conf_synch = SynchronizerConfiguration(pool_synch)
                        conf_synch_ctrl.add_channel(conf_synch)

                    ctrl_conf['synchronizer'] = conf_synch
                try:
                    synchronization = ctrl_data['synchronization']
                except KeyError:
                    synchronization = AcqSynchType.Trigger
                ctrl_conf['synchronization'] = synchronization
                user_config_ctrl['synchronization'] = synchronization

            acq_synch = None
            if external:
                ctrl_item = ExternalControllerConfiguration(ctrl)
            elif ctrl.is_timerable():
                is_software = synchronizer == 'software'
                acq_synch = AcqSynch.from_synch_type(is_software,
                                                     synchronization)
                ctrl_acq_synch[ctrl] = acq_synch
                timer = ctrl_data.get("timer")
                ctrl_conf["timer"] = timer
                monitor = ctrl_data.get("monitor")
                ctrl_conf["monitor"] = monitor
                ctrl_item = TimerableControllerConfiguration(ctrl, ctrl_conf)
            else:
                ctrl_item = ControllerConfiguration(ctrl, ctrl_conf)

            ctrl_enabled = False
            if 'channels' in ctrl_data:
                user_config_ctrl['channels'] = user_config_channel = {}
            for ch_name, ch_data in list(ctrl_data['channels'].items()):
                if external:
                    validator = TangoAttributeNameValidator()
                    full_name = ch_data.get('full_name', ch_name)
                    params = validator.getUriGroups(full_name)
                    params['pool'] = pool
                    channel = PoolExternalObject(**params)
                else:
                    channel = pool.get_element_by_full_name(ch_name)
                ch_data = self._fill_channel_data(channel, ch_data)
                user_config_channel[ch_name] = ch_data
                ch_item = ChannelConfiguration(channel, ch_data)
                ch_item.controller = ctrl_item
                ctrl_item.add_channel(ch_item)
                if ch_item.enabled:
                    if external:
                        id_ = channel.full_name
                    else:
                        id_ = channel.id
                    user_elem_ids[ch_item.index] = id_

                if ch_item.enabled:
                    ctrl_enabled = True

                if acq_synch is not None:
                    channel_acq_synch[channel] = acq_synch

            if not external and ctrl.is_timerable():
                ctrl_item.update_timer()
                ctrl_item.update_monitor()

                msg_error = ''
                if ctrl_item.timer is None:
                    timer_name = ctrl_data['timer']
                    ch_timer = pool.get_element_by_full_name(timer_name)
                    msg_error += 'Channel {0} is not present but used as ' \
                                 'timer. '.format(ch_timer.name)
                if ctrl_item.monitor is None:
                    monitor_name = ctrl_data['monitor']
                    ch_monitor = pool.get_element_by_full_name(monitor_name)
                    msg_error += 'Channel {0} is not present but used as ' \
                                 'monitor.'.format(ch_monitor.name)
                if len(msg_error) > 0:
                    raise ValueError(msg_error)

                if ctrl_item.enabled:
                    user_config_ctrl['timer'] = ctrl_item.timer.full_name
                    user_config_ctrl['monitor'] = ctrl_item.monitor.full_name
                else:
                    # TODO: Remove timer and monitor configuration
                    # timer and monitor configuration are deprecated.
                    # This is because these may be different for
                    # SoftwareTrigger, SoftwareGate or SoftwareStart
                    # synchronization type. To be decided how we will
                    # specify the new configuration e.g. chosen on the
                    # server side based on the channel's order or
                    # specified explicitly on the client side (expconf).
                    # For backwards compatibility (some clients may rely on
                    # them) we set them.
                    user_config_ctrl['timer'] = ctrl_data['timer']
                    user_config_ctrl['monitor'] = ctrl_data['monitor']

            # Update synchronizer state
            if ctrl_enabled and conf_synch is not None:
                conf_synch.enabled = True

            ctrl_item.validate()

            if external:
                other_ctrls.append(ctrl_item)
            elif ctrl.is_timerable():
                timerable_ctrls[acq_synch].append(ctrl_item)
                # Find master timer/monitor the system take the channel with
                # less index

                if not ctrl_item.enabled:
                    # Skip controllers disabled
                    pass
                elif acq_synch in (AcqSynch.SoftwareTrigger,
                                   AcqSynch.SoftwareGate):
                    if ctrl_item.timer.index < master_timer_idx_sw:
                        master_timer_sw = ctrl_item.timer
                        master_timer_idx_sw = ctrl_item.timer.index
                    if ctrl_item.monitor.index < master_monitor_idx_sw:
                        master_monitor_sw = ctrl_item.monitor
                        master_monitor_idx_sw = ctrl_item.monitor.index
                elif acq_synch == AcqSynch.SoftwareStart:
                    if ctrl_item.timer.index < master_timer_idx_sw_start:
                        master_timer_sw_start = ctrl_item.timer
                        master_timer_idx_sw_start = ctrl_item.timer.index
                    if ctrl_item.monitor.index < master_monitor_idx_sw_start:
                        master_monitor_sw_start = ctrl_item.monitor
                        master_monitor_idx_sw_start = ctrl_item.monitor.index
            elif ctrl.get_ctrl_types()[0] == ElementType.ZeroDExpChannel:
                zerod_ctrls.append(ctrl_item)

        # Update synchronizer controller states
        for conf_synch_ctrl in synch_ctrls:
            conf_synch_ctrl.update_state()

        # Fill user configuration with measurement group's timer & monitor
        # This is a backwards compatibility cause the measurement group's
        # timer & monitor are not used
        if master_timer_sw is not None:
            user_config['timer'] = master_timer_sw.full_name
        elif master_timer_sw_start is not None:
            user_config['timer'] = master_timer_sw_start.full_name
        else:  # Measurement Group with all channel synchronized by hardware
            mnt_grp_timer = cfg.get('timer')
            if mnt_grp_timer:
                user_config['timer'] = mnt_grp_timer
            else:
                # for backwards compatibility use a random monitor
                user_config['timer'] = user_config_ctrl['timer']

        if master_monitor_sw is not None:
            user_config['monitor'] = master_monitor_sw.full_name
        elif master_monitor_sw_start is not None:
            user_config['monitor'] = master_monitor_sw_start.full_name
        else:  # Measurement Group with all channel synchronized by hardware
            mnt_grp_monitor = cfg.get('monitor')
            if mnt_grp_monitor:
                user_config['monitor'] = mnt_grp_monitor
            else:
                # for backwards compatibility use a random monitor
                user_config['monitor'] = user_config_ctrl['monitor']

        # Update internals values
        self._label = label
        self._description = description
        self._timerable_ctrls = timerable_ctrls
        self._zerod_ctrls = zerod_ctrls
        self._synch_ctrls = synch_ctrls
        self._other_ctrls = other_ctrls
        self._master_timer_sw = master_timer_sw
        self._master_monitor_sw = master_monitor_sw
        self._master_timer_sw_start = master_timer_sw_start
        self._master_monitor_sw_start = master_monitor_sw_start
        self._user_config = user_config
        self._channel_acq_synch = channel_acq_synch
        self._ctrl_acq_synch = ctrl_acq_synch

        # sorted ids may not be consecutive (if a channel is disabled)
        indexes = sorted(user_elem_ids.keys())
        user_elem_ids_list = [user_elem_ids[idx] for idx in indexes]
        for conf_synch_ctrl in synch_ctrls:
            for conf_synch in conf_synch_ctrl.get_channels(enabled=True):
                user_elem_ids_list.append(conf_synch.id)
        self._parent.set_user_element_ids(user_elem_ids_list)

        self.changed = True

    def _fill_channel_data(self, channel, channel_data):
        """Fill channel default values for the given channel dictionary"""
        name = channel.name
        ctype = channel.get_type()
        full_name = channel.full_name
        # choose ndim
        ndim = None
        if ctype == ElementType.CTExpChannel:
            ndim = 0
        elif ctype == ElementType.PseudoCounter:
            ndim = 0
        elif ctype == ElementType.ZeroDExpChannel:
            ndim = 0
        elif ctype == ElementType.OneDExpChannel:
            ndim = 1
        elif ctype == ElementType.TwoDExpChannel:
            ndim = 2
        elif ctype == ElementType.External:
            config = channel.get_config()
            if config is not None:
                ndim = int(config.data_format)
        elif ctype == ElementType.IORegister:
            ndim = 0

        if ctype != ElementType.External and channel.is_referable():
            value_ref_enabled = channel_data.get('value_ref_enabled', False)
            channel_data['value_ref_enabled'] = value_ref_enabled
            value_ref_pattern = channel_data.get('value_ref_pattern', '')
            channel_data['value_ref_pattern'] = value_ref_pattern
        elif 'value_ref_enabled' in channel_data or 'value_ref_pattern' in \
                channel_data:
            if self._value_ref_compat:
                msg = 'value_ref_pattern/value_ref_enabled is deprecated ' \
                      'for non-referable channels since 3.0.3. Re-apply ' \
                      'configuration in order to upgrade.'
                self._parent.warning(msg)
                channel_data.pop('value_ref_enabled')
                channel_data.pop('value_ref_pattern')
            else:
                msg = ('The channel {} is not referable. You can not set '
                       'the enabled and/or the pattern parameters.').format(
                        name)
                raise ValueError(msg)
        # Definitively should be initialized by measurement group
        # index MUST be here already (asserting this in the following line)
        channel_data['index'] = channel_data['index']
        channel_data['name'] = channel_data.get('name', name)
        channel_data['full_name'] = channel_data.get('full_name', full_name)
        channel_data['source'] = channel.get_source()
        channel_data['enabled'] = channel_data.get('enabled', True)
        channel_data['label'] = channel_data.get('label', channel_data['name'])
        channel_data['ndim'] = ndim
        # Probably should be initialized by measurement group
        channel_data['output'] = channel_data.get('output', True)

        # Perhaps should NOT be initialized by measurement group
        channel_data['plot_type'] = channel_data.get('plot_type', PlotType.No)
        channel_data['plot_axes'] = channel_data.get('plot_axes', [])
        channel_data['conditioning'] = channel_data.get('conditioning', '')
        channel_data['normalization'] = channel_data.get('normalization',
                                                         Normalization.No)
        # TODO: think of filling other keys: data_type, data_units, nexus_path
        # shape here instead of feeling them on the Taurus extension level

        if ctype != ElementType.External:
            ctrl_name = channel.controller.full_name
            channel_data['_controller_name'] = channel_data.get(
                '_controller_name', ctrl_name)
        return channel_data


class PoolMeasurementGroup(PoolGroupElement):

    def __init__(self, **kwargs):
        self._state_lock = threading.Lock()
        self._monitor_count = None
        self._nb_starts = 1
        self._pending_starts = 0
        self._acquisition_mode = AcqMode.Timer
        self._config = MeasurementConfiguration(self)
        self._config_dirty = True
        self._moveable = None
        self._moveable_obj = None
        # by default software synchronizer initial domain is set to Position
        self._sw_synch_initial_domain = SynchDomain.Position

        self._synch_description = SynchDescription()

        kwargs['elem_type'] = ElementType.MeasurementGroup
        PoolGroupElement.__init__(self, **kwargs)
        configuration = kwargs.get("configuration")
        if configuration is None:
            user_elements = self.get_user_elements()
            configuration = build_measurement_configuration(user_elements)
        self.set_configuration_from_user(configuration)

    def _create_action_cache(self):
        acq_name = "%s.Acquisition" % self._name
        return PoolAcquisition(self, acq_name)

    def _calculate_states(self, state_info=None):
        state, status = PoolGroupElement._calculate_states(self, state_info)
        # check if software synchronizer is occupied
        synch_soft = self.acquisition._synch._synch_soft
        acq_sw = self.acquisition._sw_acq
        acq_sw_start = self.acquisition._sw_start_acq
        acq_0d = self.acquisition._0d_acq
        if (state in (State.On, State.Unknown)
            and (synch_soft.is_started()
                 or acq_sw._is_started()
                 or acq_sw_start._is_started()
                 or acq_0d._is_started())):
            state = State.Moving
            status += "\nSoftware synchronization is in progress"
        return state, status

    def on_element_changed(self, evt_src, evt_type, evt_value):
        name = evt_type.name
        if name == 'state':
            with self._state_lock:
                state, status = self._calculate_states()
                self.set_state(state, propagate=2)
                self.set_status("\n".join(status))

    def get_pool_controllers(self):
        return self.get_acquisition().get_pool_controllers()

    def get_pool_controller_by_name(self, name):
        name = name.lower()
        for ctrl in self.get_pool_controllers():
            if ctrl.name.lower() == name or ctrl.full_name.lower() == name:
                return ctrl

    def add_user_element(self, element, index=None):
        '''Override the base behavior, so the TriggerGate elements are silently
        skipped if used multiple times in the group'''
        user_elements = self._user_elements
        if element in user_elements:
            # skipping TriggerGate element if already present
            if element.get_type() is ElementType.TriggerGate:
                return
        return PoolGroupElement.add_user_element(self, element, index)

    def rename_element(self, old_name, new_name, propagate=1):
        """Rename element in the controller.

        :param old_name: old name of the element
        :type old_name: :obj:`str`
        :param new_name: new name of the element
        :type new_name: :obj:`str`
        :param propagate: 0 for not propagating, 1 to propagate,
               2 propagate with priority
        :type propagate: :obj:`int`
        """
        self._config['label'] = new_name
        self.fire_event(EventType("configuration", priority=propagate),
                        self._config)

    # -------------------------------------------------------------------------
    # configuration
    # -------------------------------------------------------------------------

    def _is_managed_element(self, element):
        element_type = element.get_type()
        return (element_type in TYPE_EXP_CHANNEL_ELEMENTS or
                element_type is ElementType.TriggerGate)

    @property
    def configuration(self):
        return self._config

    # TODO: Check if it needed
    def set_configuration(self, config=None, propagate=1):
        self._config.configuration = config
        self._config_dirty = True
        if not propagate:
            return
        self.fire_event(EventType("configuration", priority=propagate),
                        config)

    def set_configuration_from_user(self, cfg, propagate=1):
        self._config.set_configuration_from_user(cfg)
        self._config_dirty = True
        if not propagate:
            return
        self.fire_event(EventType("configuration", priority=propagate),
                        self._config.get_configuration_for_user())

    def get_user_configuration(self):
        return self._config.get_configuration_for_user()

    def get_timer(self):
        # TODO: Adapt to the new future MeasurementConfiguration API
        return self._config._master_timer

    timer = property(get_timer)

    # -------------------------------------------------------------------------
    # integration time
    # -------------------------------------------------------------------------

    def get_integration_time(self):
        integration_time = self._synch_description.active_time
        if isinstance(integration_time, float):
            return integration_time
        elif len(integration_time) == 0:
            raise Exception("The synchronization description group has not"
                            " been initialized")
        elif len(integration_time) > 1:
            raise Exception("There are more than one synchronization"
                            "description groups")

    def set_integration_time(self, integration_time, propagate=1):
        total_time = integration_time + self.latency_time
        synch = [{SynchParam.Delay: {SynchDomain.Time: 0},
                  SynchParam.Active: {SynchDomain.Time: integration_time},
                  SynchParam.Total: {SynchDomain.Time: total_time},
                  SynchParam.Repeats: 1}]
        self.set_synch_description(synch)
        if not propagate:
            return
        self.fire_event(EventType("integration_time", priority=propagate),
                        integration_time)

    integration_time = property(get_integration_time, set_integration_time,
                                doc="the current integration time")

    # -------------------------------------------------------------------------
    # monitor count
    # -------------------------------------------------------------------------

    def get_monitor_count(self):
        return self._monitor_count

    def set_monitor_count(self, monitor_count, propagate=1):
        self._monitor_count = monitor_count
        if not propagate:
            return
        self.fire_event(EventType("monitor_count", priority=propagate),
                        monitor_count)

    monitor_count = property(get_monitor_count, set_monitor_count,
                             doc="the current monitor count")

    # -------------------------------------------------------------------------
    # acquisition mode
    # -------------------------------------------------------------------------

    def get_acquisition_mode(self):
        return self._acquisition_mode

    def set_acquisition_mode(self, acquisition_mode, propagate=1):
        self._acquisition_mode = acquisition_mode
        self._config_dirty = True  # acquisition mode goes to configuration
        if not propagate:
            return
        self.fire_event(EventType("acquisition_mode", priority=propagate),
                        acquisition_mode)

    acquisition_mode = property(get_acquisition_mode, set_acquisition_mode,
                                doc="the current acquisition mode")

    # -------------------------------------------------------------------------
    # synch_description
    # -------------------------------------------------------------------------

    def get_synch_description(self):
        return self._synch_description

    def set_synch_description(self, description, propagate=1):
        self._synch_description = \
            SynchDescription(description)
        self._config_dirty = True  # acquisition mode goes to configuration
        if not propagate:
            return
        self.fire_event(EventType("synch_description",
                                  priority=propagate),
                        description)

    synch_description = property(get_synch_description, set_synch_description,
                                 doc="the current acquisition mode")

    # -------------------------------------------------------------------------
    # moveable
    # -------------------------------------------------------------------------

    def get_moveable(self):
        return self._moveable

    def set_moveable(self, moveable, propagate=1, to_fqdn=True):
        self._moveable = moveable
        if self._moveable is not None:
            if to_fqdn:
                moveable = _to_fqdn(moveable, logger=self)
            self._moveable_obj = self.pool.get_element_by_full_name(moveable)
        self.fire_event(EventType("moveable", priority=propagate),
                        moveable)

    moveable = property(get_moveable, set_moveable,
                        doc="moveable source used in synchronization "
                            "description")

    # -------------------------------------------------------------------------
    # latency time
    # -------------------------------------------------------------------------

    def get_latency_time(self):
        latency_time = 0
        pool_ctrls = self.get_pool_controllers()
        for pool_ctrl in pool_ctrls:
            if not pool_ctrl.is_timerable():
                continue
            candidate = pool_ctrl.get_ctrl_par("latency_time")
            if candidate > latency_time:
                latency_time = candidate
        return latency_time

    latency_time = property(get_latency_time,
                            doc="latency time between two consecutive "
                                "acquisitions")

    # -------------------------------------------------------------------------
    # software synchronizer initial domain
    # -------------------------------------------------------------------------

    def get_sw_synch_initial_domain(self):
        return self._sw_synch_initial_domain

    def set_sw_synch_initial_domain(self, domain):
        self._sw_synch_initial_domain = domain

    sw_synch_initial_domain = property(
        get_sw_synch_initial_domain,
        set_sw_synch_initial_domain,
        doc="software synchronizer initial domain (SynchDomain.Time "
            "or SynchDomain.Position)"
    )

    # -------------------------------------------------------------------------
    # number of starts
    # -------------------------------------------------------------------------

    def get_nb_starts(self):
        return self._nb_starts

    def set_nb_starts(self, nb_starts, propagate=1):
        self._nb_starts = nb_starts
        if not propagate:
            return
        self.fire_event(EventType("nb_starts", priority=propagate),
                        nb_starts)

    nb_starts = property(get_nb_starts, set_nb_starts,
                         doc="current number of starts")

    # -------------------------------------------------------------------------
    # acquisition
    # -------------------------------------------------------------------------

    def prepare(self):
        """Prepare for measurement.

        Delegate measurement preparation to the acquisition action.
        """
        if len(self.get_user_elements()) == 0:
            # All channels were disabled
            raise RuntimeError('all channels in measurement group '
                               'are disabled')

        if self._acquisition_mode == AcqMode.Timer:
            role = 'timer'
        elif self._acquisition_mode == AcqMode.Monitor:
            role = 'monitor'
        else:
            raise RuntimeError('acquisition mode must be either Timer '
                               'or Monitor')

        for ctrl in self._config.get_timerable_ctrls(enabled=True):
            master = getattr(ctrl, role, None)
            if master is None:
                msg = ('controller {0} does not have {1} '
                       'configured.').format(ctrl.name, role)
                raise RuntimeError(msg)
            if not master.enabled:
                msg = 'channel {0} used as {1} must be enabled'.format(
                    master.name, role)
                raise RuntimeError(msg)

        value = self._get_value()
        self._pending_starts = self.nb_starts

        kwargs = {'head': self}

        self.acquisition.prepare(self.configuration,
                                 self.acquisition_mode,
                                 value,
                                 self._synch_description,
                                 self._moveable_obj,
                                 self.sw_synch_initial_domain,
                                 self.nb_starts,
                                 **kwargs)

    def start_acquisition(self, value=None):
        """Start measurement.

        Delegate start measurement to the acquisition action.
        Provide backwards compatibility for starts without previous prepare.
        """
        if self._pending_starts == 0:
            msg = "prepare is mandatory before starting acquisition"
            raise RuntimeError(msg)
        self._aborted = False
        self._pending_starts -= 1
        if not self._simulation_mode:
            self.acquisition.run()

    def _get_value(self):
        if self._acquisition_mode is AcqMode.Timer:
            value = self.get_integration_time()
        elif self.acquisition_mode is AcqMode.Monitor:
            value = self._monitor_count
        return value

    def set_acquisition(self, acq_cache):
        self.set_action_cache(acq_cache)

    def get_acquisition(self):
        return self.get_action_cache()

    acquisition = property(get_acquisition, doc="acquisition object")

    def stop(self):
        self._pending_starts = 0
        self.acquisition._synch._synch_soft.stop()
        PoolGroupElement.stop(self)

    def abort(self):
        self._pending_starts = 0
        PoolGroupElement.abort(self)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import time
import threading
import numpy

import unittest

from sardana import State
from sardana.pool.poolcontrollers.DummyMotorController import Motion
from sardana.pool.pooldefs import SynchParam
from sardana.sardanaattribute import SardanaAttribute
from taurus.core.util.log import Logger

__all__ = ['BaseControllerTestCase', 'TriggerGateControllerTestCase',
           'PositionGenerator', 'TriggerGateReceiver']

class BaseControllerTestCase(object):
    """ Base test case for unit testing arbitrary controllers.
    This class will create a controller instance and define an axis from the
    class member attributes:
        KLASS <type> controller class
        PROPS <dict> properties of the controller
        AXIS <int> number of the axis
    """
    KLASS = None
    PROPS = {}
    AXIS = 1
    DEBUG = False

    def setUp(self):
        self.logger = Logger('BaseControllerTestCase')
        if self.DEBUG:
            self.logger.setLogLevel(Logger.Debug)

        if self.KLASS is None:
            raise Exception('Ctrl klass has not been defined')
        name = 'test_ctrl'
        self.ctrl = self.KLASS(name, self.PROPS)
        self.pre_AddDevice_hook()
        self.ctrl.AddDevice(self.AXIS)

    def tearDown(self):
        if self.ctrl is not None:
            self.ctrl.DeleteDevice(self.AXIS)

    def axisPar(self, name, value, expected_value=None):
        """ Helper for test the SetAxisPar & GetaxisPar methods
        """
        axis = self.AXIS
        if expected_value is None:
            expected_value = value
        self.ctrl.SetAxisPar(axis, name, value)
        r_value = self.ctrl.GetAxisPar(axis, name)
        msg = ('The %s value is %s, and the expected value is %s'
               % (name, r_value, expected_value))
        self.assertEqual(r_value, expected_value, msg)

    def stateOne(self, expected_state=State.On):
        """ Helper for test the stateOne method
        """
        sta, status = self.ctrl.StateOne(self.AXIS)
        msg = ('The current state of axis(%d) is %d when expected, %d'
               % (self.AXIS, sta, expected_state))
        self.assertEqual(sta, expected_state, msg)

    def start_action(self, configuration):
        """ This method set the axis parameters and pre start the axis.
        """
        for key, value in list(configuration.items()):
            self.axisPar(key, value)
        self.ctrl.SynchOne(configuration)

    def pre_AddDevice_hook(self):
        pass


class TriggerGateControllerTestCase(unittest.TestCase, BaseControllerTestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        BaseControllerTestCase.setUp(self)
        self.isAborted = False

    def tearDown(self):
        BaseControllerTestCase.tearDown(self)
        unittest.TestCase.tearDown(self)

    def post_configuration_hook(self):
        ''' Hook for post configure actions
        '''
        pass

    def post_generation_hook(self):
        ''' Hook for post generation actions
        '''
        pass

    def generation(self, configuration):
        """ Helper for test a simple generation
        """
        self.configuration = configuration
        repetitions = 0
        for group in configuration:
            repetitions += group[SynchParam.Repeats]
        # store repeats for the assers against received triggers
        self.repetitions = repetitions
        self.ctrl.SynchOne(self.AXIS, configuration)
        # execute Hook
        self.post_configuration_hook()
        # PreStartOne the axis
        self.ctrl.PreStartOne(self.AXIS)
        self.ctrl.StartOne(self.AXIS)
        while self.ctrl.StateOne(self.AXIS)[0] == State.Moving:
            time.sleep(0.001)
        self.post_generation_hook()
        state, status = self.ctrl.StateOne(self.AXIS)
        msg = ('The axis %d is not Stopped, its status is %s'
               % (self.AXIS, status))
        self.assertEqual(state, State.get('On'), msg)

    def abort(self, configuration, abort):
        """ Helper for test the abort
        """
        self.configuration = configuration
        self.ctrl.SynchOne(self.AXIS, configuration)
        self.post_configuration_hook()
        # PreStartOne the axis
        self.ctrl.PreStartOne(self.AXIS)
        self.ctrl.StartOne(self.AXIS)
        while self.ctrl.StateOne(self.AXIS)[0] == State.Moving:
            time.sleep(abort)
            self.ctrl.AbortOne(self.AXIS)
        self.isAborted = True
        self.post_generation_hook()
        state, status = self.ctrl.StateOne(self.AXIS)
        msg = ('The axis %d is not Stopped, its status is %s'
               % (self.AXIS, status))
        self.assertEqual(state, State.get('On'), msg)


class PositionGenerator(threading.Thread):
    """ It is a position generator. A Sardana Motion class is used for simulate
    the motor. The attribute value has the current user position of the motor.
    """

    def __init__(self, start_pos, end_pos, period):
        """
        :param start_pos: start position for the motion
        :param end_pos: end position for the motion
        :param period: nap time between fireevents
        :return:
        """
        threading.Thread.__init__(self)
        self.motor = Motion()
        self.motor.setMinVelocity(0)
        self.motor.setMaxVelocity(10)
        self.motor.setAccelerationTime(1)
        self.motor.setDecelerationTime(1)
        self.motor.setCurrentPosition(0)
        self._start_pos = start_pos
        self._end_pos = end_pos
        self._period = period
        self.value = SardanaAttribute(self, name='Position',
                                      initial_value=0)

    def run(self):
        """
        Start the motion and update the SardanaAttribute value with the current
        position of the motion between every nap period
        """
        self.motor.startMotion(self._start_pos, self._end_pos)
        while self.motor.isInMotion():
            value = self.motor.getCurrentUserPosition()
            self.value.set_value(value, timestamp=time.time(), propagate=1)
            time.sleep(self._period)
        value = self.motor.getCurrentUserPosition()
        self.value.set_value(value, timestamp=time.time(), propagate=1)

    def getMotor(self):
        """ Get the motion object
        """
        return self.motor

    def setStartPos(self, pos):
        """ Update start position
        """
        self._start_pos = pos
        self.value.set_value(pos)

    def setEndPos(self, pos):
        """ Update end position
        """
        self._end_pos = pos

    def setPeriod(self, time):
        """ Update the nap time
        """
        self._end_pos = time

    def add_listener(self, listener):
        self.value.add_listener(listener)

    def remove_listener(self, listener):
        self.value.remove_listener(listener)


class TriggerGateReceiver(object):
    '''Software TriggerGateReceiver which captures timestamps whenever an event
    comes. Provides useful methods for calculating the event generation
    performance
    '''
    # TODO: add more jitter measurements e.g. drift

    def __init__(self):
        self.active_events = {}
        self.passive_events = {}

    def getCount(self):
        count = len(list(self.passive_events.keys()))
        return count

    count = property(getCount)

    def event_received(self, *args, **kwargs):
        # store also a timestamp of the start event when it will be implemented
        timestamp = time.time()
        _, type_, value = args
        name = type.name
        if name == "active":
            self.active_events[value] = timestamp
        elif name == "passive":
            self.passive_events[value] = timestamp
        else:
            raise ValueError('Unknown EventType')

    def calc_characteristics(self):
        # TODO: refactor the characteristics calculation method to use numpy
        i = 0
        count = self.count
        characteristics = {}
        # there is no active event ending the last passive period, that's why
        # calculate characteristics until (count - 1)
        while i < (count - 1):
            t1 = self.active_events[i]
            t2 = self.passive_events[i]
            t3 = self.active_events[i + 1]
            active_period = t2 - t1
            passive_period = t3 - t2
            characteristics[i] = (active_period, passive_period)
            i += 1
        return characteristics

    def calc_cycletocycle(self):
        '''Calculate the cycle-to-cycle jitter characteristics: mean, std and max.
        Cycle-to-cycle jitter is a difference between a cycle period and a cycle
        period before it. To calculate one cycle-to-cycle jitter one needs
        exactly 3 active events:

        c2c_jitter_1 = cycle_2 - cycle_1
        cycle_2 = active_3 - active_2
        cycle_1 = active_2 - active_1
        '''
        i = 0
        count = self.count
        periods = []
        mean_c2c, std_c2c, max_c2c = 0, 0, 0
        # there is no active event ending the last passive period, that's why
        # calculate characteristics until (count - 1)
        while i < (count - 1):
            t1 = self.active_events[i]
            t2 = self.active_events[i + 1]
            period = t2 - t1
            periods.append(period)
            i += 1
        if len(periods) > 0:
            periods_array = numpy.array(periods)
            print(periods_array)
            c2c = numpy.diff(periods_array)
            mean_c2c = c2c.mean()
            std_c2c = c2c.std()
            max_c2c = c2c.max()
        return mean_c2c, std_c2c, max_c2c



#!/usr/bin/env python
import PyTango

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__all__ = ["TriggerGate", "TriggerGateClass"]

__docformat__ = 'restructuredtext'

import sys
import time

from PyTango import DevFailed, DevVoid, DevLong, DevState, AttrQuality, \
    DevString, Except, SCALAR, ErrSeverity, READ

from taurus.core.util.log import DebugIt

from sardana import State, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.tango.core.util import to_tango_type_format, exception_str

from sardana.tango.pool.PoolDevice import PoolElementDevice, \
    PoolElementDeviceClass


class TriggerGate(PoolElementDevice):

    def __init__(self, dclass, name):
        PoolElementDevice.__init__(self, dclass, name)

    def init(self, name):
        PoolElementDevice.init(self, name)

    def _is_allowed(self, req_type):
        return PoolElementDevice._is_allowed(self, req_type)

    def get_tg(self):
        return self.element

    def set_tg(self, tg):
        self.element = tg

    tg = property(get_tg, set_tg)

    @DebugIt()
    def delete_device(self):
        PoolElementDevice.delete_device(self)
        tg = self.tg
        if tg is not None:
            tg.remove_listener(self.on_tg_changed)

    @DebugIt()
    def init_device(self):
        PoolElementDevice.init_device(self)

        tg = self.tg
        if tg is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.tg = tg = \
                self.pool.create_element(type="TriggerGate",
                                         name=name, full_name=full_name, id=self.Id, axis=self.Axis,
                                         ctrl_id=self.Ctrl_id)

        tg.add_listener(self.on_tg_changed)
        self.set_state(DevState.ON)

    def on_tg_changed(self, event_source, event_type, event_value):
        try:
            self._on_tg_changed(event_source, event_type, event_value)
        except not DevFailed:
            msg = 'Error occurred "on_tg_changed(%s.%s): %s"'
            exc_info = sys.exc_info()
            self.error(msg, event_type.name,
                       exception_str(*exc_info[:2]))
            self.debug("Details", exc_info=exc_info)

    def _on_tg_changed(self, event_source, event_type, event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name.lower()

        try:
            attr = self.get_attribute_by_name(name)
        except DevFailed:
            return

        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        value, w_value, error = None, None, None

        if name == "state":
            value = self.calculate_tango_state(event_value)
        elif name == "status":
            value = self.calculate_tango_status(event_value)
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                else:
                    value = event_value.value
                timestamp = event_value.timestamp
            else:
                value = event_value

        self.set_attribute(attr, value=value, w_value=w_value,
                           timestamp=timestamp, quality=quality,
                           priority=priority, error=error, synch=False)


class TriggerGateClass(PoolElementDeviceClass):

    def _get_class_properties(self):
        ret = PoolElementDeviceClass._get_class_properties(self)
        ret['Description'] = "Trigger/Gate device class"
        ret['InheritedFrom'].insert(0, 'PoolElementDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

__all__ = ['dummyMeasurementGroupConf01',
           'dummyPoolCTCtrlConf01', 'dummyPoolCTCtrlConf02',
           'TangoAttributeCTCtrlConf01', 'TangoAttributeCTCtrlConf02',
           'dummyPoolTGCtrlConf01', 'dummyPoolTGCtrlConf02',
           'dummyCounterTimerConf01', 'dummyCounterTimerConf02',
           'dummyCounterTimerConf03', 'dummyCounterTimerConf04',
           'dummyTriggerGateConf01', 'dummyTriggerGateConf02',
           'dummyTriggerGateConf03', 'dummyTriggerGateConf04',
           'dummyPoolMotorCtrlConf01', 'dummyMotorConf01', 'dummyMotorConf02']


# Pool Measurement Group
'''Minimum configuration to create a Pool MeasurementGroup'''
dummyMeasurementGroupConf01 = {'full_name': 'mntgrp01',
                               'id': None,
                               'name': '',
                               'pool': None,
                               'user_elements': [2]}


# Pool Ctrls
'''Minimum configuration to create a Pool CounterTimer controller'''
dummyPoolCTCtrlConf01 = {'class_info': None,
                         'full_name': 'dctctrl01',
                         'id': 1,
                         'klass': 'DummyCounterTimerController',
                         'lib_info': None,
                         'library': 'DummyCounterTimerController.py',
                         'name': '',
                         'pool': None,
                         'properties': {},
                         'role_ids': '',
                         'type': 'CTExpChannel'}

'''Second minimum configuration to create a Pool CounterTimer controller'''
dummyPoolCTCtrlConf02 = {'class_info': None,
                         'full_name': 'dctctrl02',
                         'id': 2,
                         'klass': 'DummyCounterTimerController',
                         'lib_info': None,
                         'library': 'DummyCounterTimerController.py',
                         'name': '',
                         'pool': None,
                         'properties': {},
                         'role_ids': '',
                         'type': 'CTExpChannel'}

'''Minimum configuration to create a TangoAttributeCTcontroller'''
TangoAttributeCTCtrlConf01 = {'class_info': None,
                              'ctrl_info': '',
                              'full_name': 'tctctrl01',
                              'id': 1,
                              'klass': 'TangoAttrCTController',
                              'lib_info': None,
                              #'library': 'TangoAttrCTCtrl.py',
                              'library': 'TangoAttrCounterTimerController.py',
                              'name': '',
                              'pool': None,
                              'properties': {},
                              'role_ids': '',
                              'type': 'CTExpChannel'}

'''Second minimum configuration to create a TangoAttributeCTcontroller'''
TangoAttributeCTCtrlConf02 = {'class_info': None,
                              'ctrl_info': '',
                              'full_name': 'tctctrl02',
                              'id': 2,
                              'klass': 'TangoAttrCTController',
                              'lib_info': None,
                              #'library': 'TangoAttrCTCtrl.py',
                              'library': 'TangoAttrCounterTimerController.py',
                              'name': '',
                              'pool': None,
                              'properties': {},
                              'role_ids': '',
                              'type': 'CTExpChannel'}

'''Minimum configuration to create a Pool TriggerGate controller'''
dummyPoolTGCtrlConf01 = {'class_info': None,
                         'full_name': 'dtgctrl01',
                         'id': 1,
                         'klass': 'DummyTriggerGateController',
                         'lib_info': None,
                         'library': 'DummyTriggerGateController.py',
                         'name': '',
                         'pool': None,
                         'properties': {},
                         'role_ids': '',
                         'type': 'TGChannel'}

'''Minimum configuration to create a Pool TriggerGate controller'''
dummyPoolTGCtrlConf02 = {'class_info': None,
                         'full_name': 'dtgctrl01',
                         'id': 2,
                         'klass': 'DummyTriggerGateController',
                         'lib_info': None,
                         'library': 'DummyTriggerGateController.py',
                         'name': '',
                         'pool': None,
                         'properties': {},
                         'role_ids': '',
                         'type': 'TGChannel'}

'''Minimum configuration to create a Pool Motor controller'''
dummyPoolMotorCtrlConf01 = {'class_info': None,
                            'full_name': 'dmotctrl01',
                            'id': 1,
                            'klass': 'DummyMotorController',
                            'lib_info': None,
                            'library': 'DummyMotorController.py',
                            'name': '',
                            'pool': None,
                            'properties': {},
                            'role_ids': '',
                            'type': 'Motor'}

# Pool Elements
'''Minimum configuration to create a Pool CounterTimer'''
dummyCounterTimerConf01 = {'axis': 1,
                           'ctrl': None,
                           'full_name': 'dct01',
                           'id': 2,
                           'name': '',
                           'pool': None}

'''Second minimum configuration to create a Pool CounterTimer'''
dummyCounterTimerConf02 = {'axis': 2,
                           'ctrl': None,
                           'full_name': 'dct02',
                           'id': 3,
                           'name': '',
                           'pool': None}

'''Third minimum configuration to create a Pool CounterTimer'''
dummyCounterTimerConf03 = {'axis': 1,
                           'ctrl': None,
                           'full_name': 'dct03',
                           'id': 2,
                           'name': '',
                           'pool': None}

'''Fourth minimum configuration to create a Pool CounterTimer'''
dummyCounterTimerConf04 = {'axis': 2,
                           'ctrl': None,
                           'full_name': 'dct04',
                           'id': 3,
                           'name': '',
                           'pool': None}

'''Minimum configuration to create a Pool TriggerGate'''
dummyTriggerGateConf01 = {'axis': 1,
                          'ctrl': None,
                          'full_name': 'dtg01',
                          'id': 2,
                          'name': '',
                          'pool': None}

'''Second minimum configuration to create a Pool TriggerGate'''
dummyTriggerGateConf02 = {'axis': 2,
                          'ctrl': None,
                          'full_name': 'dtg02',
                          'id': 3,
                          'name': '',
                          'pool': None}

'''Third minimum configuration to create a Pool TriggerGate'''
dummyTriggerGateConf03 = {'axis': 3,
                          'ctrl': None,
                          'full_name': 'dtg03',
                          'id': 4,
                          'name': '',
                          'pool': None}

'''Fourth minimum configuration to create a Pool TriggerGate'''
dummyTriggerGateConf04 = {'axis': 4,
                          'ctrl': None,
                          'full_name': 'dtg04',
                          'id': 5,
                          'name': '',
                          'pool': None}

'''Minimum configuration to create a Pool Motor'''
dummyMotorConf01 = {'axis': 1,
                    'ctrl': None,
                    'full_name': 'dmot01',
                    'id': 1,
                    'name': '',
                    'pool': None}

'''Minimum configuration to create a Pool Motor'''
dummyMotorConf02 = {'axis': 2,
                    'ctrl': None,
                    'full_name': 'dmot02',
                    'id': 1,
                    'name': '',
                    'pool': None}



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import PyTango
import taurus

from sardana.tango.pool.test import BasePoolTestCase


__all__ = ['SarTestTestCase']


def _cleanup_device(dev_name):
    factory = taurus.Factory()
    device = taurus.Device(dev_name)
    # tango_alias_devs contains any names in which we have referred
    # to the device, could be alias, short name, etc. pop all of them
    for k, v in list(factory.tango_alias_devs.items()):
        if v is device:
            factory.tango_alias_devs.pop(k)
    full_name = device.getFullName()
    if full_name in factory.tango_devs:
        factory.tango_devs.pop(full_name)
    device.cleanUp()


class SarTestTestCase(BasePoolTestCase):
    """ Base class to setup sardana test environment.
        It creates the controllers defined in cls_list and pseudo_cls_list
        with the given 'n' elements.

        - cls_list is a list of tuples: (ctrl_class, prefix, subfix, num_elem)

        The ctrls name and elements name will be hardcode following
        the next structure:

        - The ctrl_name will be prefix + _ctrl_ + postfix.
        - The elem_name will be prefix + _ + postfix + _ + axis
    """
    # TODO: Formating PEP8
    cls_list = [
        ('Motor', 'DummyMotorController', 'DummyMotorController', '_test_mt', '1', 5),
        ('CTExpChannel', 'DummyCounterTimerController',
         'DummyCounterTimerController', '_test_ct', '1', 5),
        ('CTExpChannel', 'DummyCounterTimerController',
         'DummyCounterTimerController', '_test_ct', '2', 5),
        ('ZeroDExpChannel', 'DummyZeroDController',
         'DummyZeroDController', '_test_0d', '1', 5),
        ('ZeroDExpChannel', 'DummyZeroDController',
         'DummyZeroDController', '_test_0d', '2', 5),
        ('OneDExpChannel', 'DummyOneDController',
         'DummyOneDController', '_test_1d', '1', 5),
        ('OneDExpChannel', 'DummyOneDController',
         'DummyOneDController', '_test_1d', '2', 5),
        ('TwoDExpChannel', 'DummyTwoDController',
         'DummyTwoDController', '_test_2d', '1', 5),
        ('TwoDExpChannel', 'DummyTwoDController',
         'DummyTwoDController', '_test_2d', '2', 5),
        ('TriggerGate', 'DummyTriggerGateController',
         'DummyTriggerGateController', '_test_tg', '1', 5)
    ]

    pseudo_cls_list = [
        ('PseudoCounter', 'IoverI0',
         'IoverI0', '_test_pc', '1', "I=_test_ct_1_2", "I0=_test_ct_1_1",
         "IoverI0=_test_pc_1_1")
    ]

    def setUp(self, pool_properties=None):
        BasePoolTestCase.setUp(self, pool_properties)

        # due to problems with factory cleanup in Taurus 3
        # we must skip these tests temporarily, whenever Taurus 3 support will
        # be dropped, re-enable them
        if taurus.core.release.version_info[0] < 4:
            self.skipTest("Taurus 3 has problems with factory cleanup")


        self.ctrl_list = []
        self.elem_list = []
        try:
            # physical controllers and elements
            for sar_type, lib, cls, prefix, postfix, nelem in self.cls_list:
                # create controller
                ctrl_name = prefix + "_ctrl_%s" % (postfix)
                try:
                    self.pool.CreateController([sar_type, lib, cls, ctrl_name])
                    if cls in ("DummyCounterTimerController",
                               "DummyTwoDController"):
                        ctrl = PyTango.DeviceProxy(ctrl_name)
                        # use the first trigger/gate element by default
                        ctrl.write_attribute("Synchronizer", "_test_tg_1_1")
                except Exception as e:
                    print(e)
                    msg = 'Impossible to create ctrl: "%s"' % (ctrl_name)
                    raise Exception('Aborting SartestTestCase: %s' % (msg))
                self.ctrl_list.append(ctrl_name)
                # create elements
                for axis in range(1, nelem + 1):
                    elem_name = prefix + "_" + postfix + '_%s' % (axis)
                    try:
                        self.pool.createElement(
                            [sar_type, ctrl_name, str(axis), elem_name])
                    except Exception as e:
                        print(e)
                        msg = 'Impossible to create element: "%s"' % (
                            elem_name)
                        raise Exception('Aborting SartestTestCase: %s' % (msg))
                    self.elem_list.append(elem_name)
            # pseudo controllers and elements
            for pseudo in self.pseudo_cls_list:
                sar_type, lib, cls, prefix, postfix = pseudo[0:5]
                roles = pseudo[5:]
                # Create controller
                ctrl_name = prefix + "_ctrl_%s" % (postfix)
                argin = [sar_type, lib, cls, ctrl_name]
                argin.extend(roles)
                try:
                    self.pool.CreateController(argin)
                except Exception as e:
                    print(e)
                    msg = 'Impossible to create ctrl: "%s"' % (ctrl_name)
                    raise Exception('Aborting SartestTestCase: %s' % (msg))
                self.ctrl_list.append(ctrl_name)
                for role in roles:
                    elem = role.split("=")[1]
                    if elem not in self.elem_list:
                        self.elem_list.append(elem)
        except Exception as e:
            # force tearDown in order to eliminate the Pool
            BasePoolTestCase.tearDown(self)
            print(e)

    def tearDown(self):
        """Remove the elements and the controllers
        """
        dirty_elems = []
        dirty_ctrls = []
        f = taurus.Factory()
        for elem_name in self.elem_list:
            # Cleanup eventual taurus devices. This is especially important
            # if the sardana-taurus extensions are in use since this
            # devices are created and destroyed within the testsuite.
            # Persisting taurus device may react on API_EventTimeouts, enabled
            # polling, etc.
            if elem_name in f.tango_alias_devs:
                _cleanup_device(elem_name)
            try:
                self.pool.DeleteElement(elem_name)
            except Exception as e:
                print(e)
                dirty_elems.append(elem_name)

        for ctrl_name in self.ctrl_list:
            # Cleanup eventual taurus devices. This is especially important
            # if the sardana-taurus extensions are in use since this
            # devices are created and destroyed within the testsuite.
            # Persisting taurus device may react on API_EventTimeouts, enabled
            # polling, etc.
            if ctrl_name in f.tango_alias_devs:
                _cleanup_device(ctrl_name)
            try:
                self.pool.DeleteElement(ctrl_name)
            except:
                dirty_ctrls.append(ctrl_name)

        _cleanup_device(self.pool_name)

        BasePoolTestCase.tearDown(self)

        if dirty_elems or dirty_ctrls:
            msg = "Cleanup failed. Database may be left dirty." + \
                "\n\tCtrls : %s\n\tElems : %s" % (dirty_ctrls, dirty_elems)
            raise Exception(msg)


if __name__ == "__main__":
    stc = SarTestTestCase()
    stc.setUp()
    import time
    time.sleep(15)
    stc.tearDown()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

__docformat__ = 'restructuredtext'

import sys
from taurus.external.qt import Qt
import taurus.core
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.input import TaurusValueLineEdit

from taurus.external.qt import QtCore, QtGui

from taurus.qt.qtgui.util.ui import UILoadable


@UILoadable(with_ui="_ui")
class ReflectionsEditor(TaurusWidget):

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode=designMode)

        self.loadUi(filename="reflectionseditor.ui")

        self._ui.ApplyButton.clicked.connect(self.apply)
        self._ui.ClearButton.clicked.connect(self.clear)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'reflectionseditor'
        ret['group'] = 'Taurus Containers'
        ret['container'] = ':/designer/frame.png'
        ret['container'] = False
        return ret

    def setModel(self, model):
        if model is not None:
            self.device = taurus.Device(model)

        xhkl = []
        xangles = []
        xlabels = []
        for i in range(0, 6):
            xhkl.append(70 + 80 * i)
            xangles.append(310 + 80 * i)
            xlabels.append(340 + 90 * i)
        ybasis = 20

        reflections = self.device.reflectionlist

        angle_names = []
        self.angle_labels = []
        self.hkl_values = []
        self.angle_values = []
        self.index_values = []

        # Find number of real motors
        self.nb_angles = len(self.device.motorlist)

        if reflections is not None:
            self.nb_reflections = len(reflections)
        else:
            self.nb_reflections = 0

        try:
            angle_names = self.device.motorroles
        except:  # Only for compatibility
            if self.nb_angles == 4:
                angle_names.append("omega")
                angle_names.append("chi")
                angle_names.append("phi")
                angle_names.append("tth")
            elif self.nb_angles == 6:
                angle_names.append("mu")
                angle_names.append("th")
                angle_names.append("chi")
                angle_names.append("phi")
                angle_names.append("gamma")
                angle_names.append("delta")

        for jref in range(0, 10):
            self.index_values.append(QtGui.QLineEdit(self))
            self.index_values[jref].setLayoutDirection(QtCore.Qt.RightToLeft)
            self.index_values[jref].setGeometry(
                QtCore.QRect(20, ybasis + 30 * (jref + 1), 51, 27))
            object_name = "indexvalue" + str(jref)
            self.index_values[jref].setObjectName(object_name)

        for i in range(0, 3):
            self.hkl_values.append([])
            for jref in range(0, 10):
                self.hkl_values[i].append(QtGui.QLineEdit(self))
                self.hkl_values[i][jref].setLayoutDirection(
                    QtCore.Qt.RightToLeft)
                self.hkl_values[i][jref].setGeometry(
                    QtCore.QRect(xhkl[i], ybasis + 30 * (jref + 1), 81, 27))
                object_name = "hklvalue" + str(i) + "_" + str(jref)
                self.hkl_values[i][jref].setObjectName(object_name)

        for i in range(0, self.nb_angles):
            self.angle_labels.append(QtGui.QLabel(self))
            self.angle_labels[i].setGeometry(
                QtCore.QRect(xangles[i], ybasis, 70, 20))
            self.angle_labels[i].setLayoutDirection(QtCore.Qt.RightToLeft)
            object_name = "anglelabel" + str(i)
            self.angle_labels[i].setObjectName(object_name)
            self.angle_labels[i].setText(QtGui.QApplication.translate(
                "Form", angle_names[i], None))
            self.angle_values.append([])
            for jref in range(0, 10):
                self.angle_values[i].append(QtGui.QLineEdit(self))
                self.angle_values[i][jref].setLayoutDirection(
                    QtCore.Qt.RightToLeft)
                self.angle_values[i][jref].setGeometry(QtCore.QRect(
                    xangles[i], ybasis + 30 * (jref + 1), 81, 27))
                object_name = "anglevalue" + str(i) + "_" + str(jref)
                self.angle_values[i][jref].setObjectName(object_name)

        for jref in range(0, self.nb_reflections):
            ref = reflections[jref]
            # Fill index
            self.index_values[jref].setText(str(jref))
            # Fill hkl values
            for i in range(0, 3):
                self.hkl_values[i][jref].setText(
                    ("%12.4f" % ref[i + 1]).strip())
            # Fill the angle values
            for i in range(0, self.nb_angles):
                self.angle_values[i][jref].setText(
                    ("%12.4f" % ref[i + 6]).strip())

    def apply(self):
        # Get the values for the new reflections
        hklnew = []
        anglesnew = []
        indexnew = []
        iref_hkl = []
        iref_angles = []
        for jref in range(0, 10):
            hklnew.append([])
            anglesnew.append([])
            try:
                indexnew.append(int(self.index_values[jref].text()))
            except:
                indexnew.append(-1)
            icount = 0
            for ihkl in range(0, 3):
                try:
                    hklnew[jref].append(
                        float(self.hkl_values[ihkl][jref].text()))
                    icount = icount + 1
                except:
                    hklnew[jref].append('')
            iref_hkl.append(icount)
            icount = 0
            for iangle in range(0, self.nb_angles):
                try:
                    anglesnew[jref].append(
                        float(self.angle_values[iangle][jref].text()))
                    icount = icount + 1
                except:
                    anglesnew[jref].append('')
            iref_angles.append(icount)

        # Remove all reflections
        if self.device.reflectionlist is not None:
            self.nb_reflections = len(self.device.reflectionlist)
        else:
            self.nb_reflections = 0
        for i in range(0, self.nb_reflections):
            # The index is reset after removing, so we remove always the first
            # one
            self.device.write_attribute("removereflection", 0)

        # Create reflections (always with current angles attached)
        for jref in range(0, 10):
            for iref in range(0, 10):
                if indexnew[jref] == iref:
                    self.index_values[jref].setText(str(jref))
                    if iref_hkl[iref] == 3:
                        self.device.write_attribute(
                            "addreflection", hklnew[iref])
                # Set the angles if given
                    if iref_angles[iref] == self.nb_angles:
                        cmd = []
                        cmd.append(iref)
                        for i in range(0, self.nb_angles):
                            cmd.append(anglesnew[iref][i])
                        self.device.write_attribute(
                            "adjustanglestoreflection", cmd)

        self.clear()

    def clear(self):
        reflections = self.device.reflectionlist

        # Clean the values
        for jref in range(0, 10):
            # Fill hkl values
            for i in range(0, 3):
                self.hkl_values[i][jref].setText('')
            # Fill the angle values
            for i in range(0, self.nb_angles):
                self.angle_values[i][jref].setText('')

        # Add the reflections
        if reflections is not None:
            self.nb_reflections = len(reflections)
            for jref in range(0, len(reflections)):
                # Fill the index
                self.index_values[jref].setText(str(jref))
                ref = reflections[jref]
                # Fill hkl values
                for i in range(0, 3):
                    self.hkl_values[i][jref].setText(
                        ("%12.4f" % ref[i + 1]).strip())
                # Fill the angle values
                for i in range(0, self.nb_angles):
                    self.angle_values[i][jref].setText(
                        ("%12.4f" % ref[i + 6]).strip())
        else:
            self.nb_reflections = 0


def main():
    app = Qt.QApplication(sys.argv)
    w = ReflectionsEditor()
    w.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################


"""This module is part of the Python Pool library. It defines the classes
for the synchronization"""

__all__ = ["PoolSynchronization", "SynchDescription", "TGChannel"]

import time
import threading
from functools import partial
from taurus.core.util.log import DebugIt
from sardana import State
from sardana.sardanathreadpool import get_thread_pool
from sardana.pool.pooldefs import SynchDomain, SynchParam
from sardana.pool.poolaction import ActionContext, PoolActionItem, PoolAction
from sardana.util.funcgenerator import FunctionGenerator

# The purpose of this class was inspired on the CTAcquisition concept


class TGChannel(PoolActionItem):
    """An item involved in the trigger/gate generation.
    Maps directly to a trigger object

    .. note::
        The TGChannel class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, trigger_gate, info=None):
        PoolActionItem.__init__(self, trigger_gate)
        if info:
            self.__dict__.update(info)

    def __getattr__(self, name):
        return getattr(self.element, name)


class SynchDescription(list):
    """Synchronization description. It is composed from groups - repetitions
    of equidistant synchronization events. Each group is described by
    :class:`~sardana.pool.pooldefs.SynchParam` parameters which may have
    values in :class:`~sardana.pool.pooldefs.SynchDomain` domains.
    """

    @property
    def repetitions(self):
        repetitions = 0
        for group in self:
            repetitions += group[SynchParam.Repeats]
        return repetitions

    @property
    def active_time(self):
        return self._get_param(SynchParam.Active)

    @property
    def total_time(self):
        return self._get_param(SynchParam.Total)

    @property
    def passive_time(self):
        return self.total_time - self.active_time

    def _get_param(self, param, domain=SynchDomain.Time):
        """
        Extract parameter from synchronization description and its groups. If
        there is only one group in the synchronization description then
        returns float with the value. Otherwise a list of floats with
        different values.

        :param param: parameter type
        :type param: :class:`~sardana.pool.pooldefs.SynchParam`
        :param domain: domain
        :type param: :class:`~sardana.pool.pooldefs.SynchDomain`
        :return: parameter value(s)
        :rtype float or [float]
        """

        if len(self) == 1:
            return self[0][param][domain]

        values = []
        for group in self:
            value = group[param][domain]
            repeats = group[SynchParam.Repeats]
            values += [value] * repeats
        return values


class PoolSynchronization(PoolAction):
    """Synchronization action.

    It coordinates trigger/gate elements and software synchronizer.

    .. todo: Think of moving the ready/busy mechanism to PoolAction
    """

    def __init__(self, main_element, name="Synchronization"):
        PoolAction.__init__(self, main_element, name)
        # Even if rest of Sardana is using "." in logger names use "-" as
        # sepator. This is in order to avoid confusion about the logger
        # hierary - by default python logging use "." to indicate loggers'
        # hirarchy in case parent-children relation is established between the
        # loggers.
        # TODO: review if it is possible in Sardana to use a common separator.
        soft_synch_name = main_element.name + "-SoftSynch"
        self._synch_soft = FunctionGenerator(name=soft_synch_name)
        self._listener = None
        self._ready = threading.Event()
        self._ready.set()

    def _is_ready(self):
        return self._ready.is_set()

    def _wait(self, timeout=None):
        return self._ready.wait(timeout)

    def _set_ready(self, _=None):
        self._ready.set()

    def _is_busy(self):
        return not self._ready.is_set()

    def _set_busy(self):
        self._ready.clear()

    def add_listener(self, listener):
        self._listener = listener

    def start_action(self, ctrls, synch_description, moveable=None,
                     sw_synch_initial_domain=None, *args, **kwargs):
        """Start synchronization action.

        :param ctrls: list of enabled trigger/gate controllers
        :type ctrls: list
        :param synch_description: synchronization description
        :type synch_description:
         :class:`~sardana.pool.poolsynchronization.SynchDescription`
        :param moveable: (optional) moveable object used as the
         synchronization source in the Position domain
        :type moveable: :class:`~sardna.pool.poolmotor.PoolMotor` or
         :class:`~sardana.pool.poolpseudomotor.PoolPseudoMotor`
        :param sw_synch_initial_domain: (optional) - initial domain for
         software synchronizer, can be either
         :obj:`~sardana.pool.pooldefs.SynchDomain.Time` or
         :obj:`~sardana.pool.pooldefs.SynchDomain.Position`
        """
        with ActionContext(self):

            # loads synchronization description
            for ctrl in ctrls:
                pool_ctrl = ctrl.element
                pool_ctrl.ctrl.PreSynchAll()
                for channel in ctrl.get_channels(enabled=True):
                    axis = channel.axis
                    ret = pool_ctrl.ctrl.PreSynchOne(axis, synch_description)
                    if not ret:
                        msg = ("%s.PreSynchOne(%d) returns False" %
                               (ctrl.name, axis))
                        raise Exception(msg)
                    pool_ctrl.ctrl.SynchOne(axis, synch_description)
                pool_ctrl.ctrl.SynchAll()

            # attaching listener (usually acquisition action)
            # to the software trigger gate generator
            if self._listener is not None:
                if sw_synch_initial_domain is not None:
                    self._synch_soft.initial_domain = sw_synch_initial_domain
                self._synch_soft.set_configuration(synch_description)
                self._synch_soft.add_listener(self._listener)
                remove_acq_listener = partial(self._synch_soft.remove_listener,
                                              self._listener)
                self.add_finish_hook(remove_acq_listener, False)
                self._synch_soft.add_listener(
                    self.main_element.on_element_changed)
                remove_mg_listener = partial(self._synch_soft.remove_listener,
                                             self.main_element)
                self.add_finish_hook(remove_mg_listener, False)
            # subscribing to the position change events to generate events
            # in position domain
            if moveable is not None:
                position = moveable.get_position_attribute()
                position.add_listener(self._synch_soft)
                remove_pos_listener = partial(position.remove_listener,
                                              self._synch_soft)
                self.add_finish_hook(remove_pos_listener, False)

            # start software synchronizer
            if self._listener is not None:
                self._synch_soft.start()
                get_thread_pool().add(self._synch_soft.run)

            # PreStartAll on all controllers
            for ctrl in ctrls:
                pool_ctrl = ctrl.element
                pool_ctrl.ctrl.PreStartAll()

            # PreStartOne & StartOne on all elements
            for ctrl in ctrls:
                pool_ctrl = ctrl.element
                for channel in ctrl.get_channels(enabled=True):
                    axis = channel.axis
                    ret = pool_ctrl.ctrl.PreStartOne(axis)
                    if not ret:
                        raise Exception("%s.PreStartOne(%d) returns False"
                                        % (pool_ctrl.name, axis))
                    pool_ctrl.ctrl.StartOne(axis)

            # set the state of all elements to inform their listeners
            self._channels = []
            for ctrl in ctrls:
                for channel in ctrl.get_channels(enabled=True):
                    channel.set_state(State.Moving, propagate=2)
                    self._channels.append(channel)

            # StartAll on all controllers
            for ctrl in ctrls:
                pool_ctrl = ctrl.element
                pool_ctrl.ctrl.StartAll()

    def is_triggering(self, states):
        """Determines if we are synchronizing or not based on the states
        returned by the controller(s) and the software synchronizer.

        :param states: a map containing state information as returned by
                       read_state_info: ((state, status), exception_error)
        :type states: dict<PoolElement, tuple(tuple(int, str), str))
        :return: returns True if is triggering or False otherwise
        :rtype: bool
        """
        for elem in states:
            state_info_idx = 0
            state_idx = 0
            state_tggate = states[elem][state_info_idx][state_idx]
            if self._is_in_action(state_tggate):
                return True
        return False

    @DebugIt()
    def action_loop(self):
        """action_loop method
        """
        states = {}
        for channel in self._channels:
            element = channel.element
            states[element] = None

        # Triggering loop
        # TODO: make nap configurable (see motion or acquisition loops)
        nap = 0.01
        while True:
            self.read_state_info(ret=states)
            if not self.is_triggering(states):
                break
            time.sleep(nap)

        # Set element states after ending the triggering
        for element, state_info in list(states.items()):
            with element:
                element.clear_operation()
                state_info = element._from_ctrl_state_info(state_info)
                element.set_state_info(state_info, propagate=2)

        # wait for software synchronizer to finish
        if self._listener is not None:
            while True:
                if not self._synch_soft.is_started():
                    break
                time.sleep(0.01)



import PyTango
import ZeroDController
import time


class ElecMeterController(ZeroDController.ZeroDController):
    "This class is the Tango Sardana Zero D controller for an Electrometer"

    ctrl_extra_attributes = {'Py0D_extra_1': {'Type': 'PyTango.DevDouble', 'R/W Type': 'PyTango.READ_WRITE'},
                             'Py0D_extra_2': {'Type': 'PyTango.DevLong', 'R/W Type': 'PyTango.READ'},
                             'Py0D_extra_3': {'Type': 'PyTango.DevBoolean', 'R/W Type': 'PyTango.READ'}}

#    class_prop = {'CtrlDevName':{'Type':'PyTango.DevString','Description':'The ctrl simulator Tango device name'}}

    MaxDevice = 1

    def __init__(self, inst, props):
        ZeroDController.ZeroDController.__init__(self, inst, props)
        print "PYTHON -> ZeroDController ctor for instance", inst
#	raise NameError,"Ouuups"

        self.ct_name = "ZeroDController/" + self.inst_name
#        self.simu_ctrl = None
#        self.simu_ctrl = PyTango.DeviceProxy(self.CtrlDevName)
#	self.started = False

#        try:
#            self.simu_ctrl.ping()
#        except:
#            self.simu_ctrl = None
#            raise

    def AddDevice(self, ind):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In AddDevice method for index", ind
#        raise RuntimeError,"Hola la la"

    def DeleteDevice(self, ind):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In DeleteDevice method for index", ind

    def StateOne(self, ind):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In StateOne method for index", ind
#        if self.simu_ctrl != None:
#	    if self.started == True:
#	        now = time.time()
#		delta_t = now - self.start_time
#		print "delta_t =",delta_t
#		if delta_t > 2.0:
#		     self.simu_ctrl.command_inout("Stop",self.wantedCT[0])
#		     self.started = False
#            sta = self.simu_ctrl.command_inout("GetCounterState",ind)
#            print "State in controller =",sta
#            tup = (sta,"Status error string from controller")
#        else:
#            raise RuntimeError,"Ctrl Tango's proxy null!!!"
        tup = (PyTango.DevState.ON, "Status error string from controller")
        return tup

    def PreReadAll(self):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In PreReadAll method"

    def PreReadOne(self, ind):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In PreReadOne method for index", ind

    def ReadAll(self):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In ReadAll method"

    def ReadOne(self, ind):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In ReadOne method for index", ind
#	if self.simu_ctrl != None:
#            return self.simu_ctrl.command_inout("GetCounterValue",ind)
#	else:
#	    raise RuntimeError,"Ctrl Tango's proxy null!!!"
        return 1.234

    def GetExtraAttributePar(self, ind, name):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In GetExtraFeaturePar method for index", ind, " name=", name
        if name == "Py0D_extra_1":
            return 88.99

        if name == "Py0D_extra_2":
            return 33
        return 2233

    def SetExtraAttributePar(self, ind, name, value):
        print "PYTHON -> ZeroDController/", self.inst_name, ": In SetExtraFeaturePar method for index", ind, " name=", name, " value=", value

    def SendToCtrl(self, in_data):
        print "Received value =", in_data
        return "Adios"

    def __del__(self):
        print "PYTHON -> ZeroDController/", self.inst_name, ": Aarrrrrg, I am dying"


if __name__ == "__main__":
    obj = ZeroDController('test')
#    obj.AddDevice(2)
#    obj.DeleteDevice(2)
